import logging
from typing import Callable, Dict, TYPE_CHECKING, Hashable, Union

import pandas as pd

from . import FeatureGenerator, MultiFeatureGenerator
from ..data_transformation import DFTNormalisation, DFTOneHotEncoder
from ..util.string import list_string

if TYPE_CHECKING:
    pass

log = logging.getLogger(__name__)


class FeatureGeneratorRegistry:
    """
    Represents a registry for (named) feature generator factories
    """
    def __init__(self, use_singletons: bool = False):
        """
        :param use_singletons: if True, internally maintain feature generator singletons, such that there is at most one
            instance for each name/key
        """
        self._feature_generator_factories: Dict[Hashable, Callable[[], FeatureGenerator]] = {}
        self._feature_generator_singletons: Dict[Hashable, FeatureGenerator] = {}
        self._use_singletons = use_singletons

    @property
    def available_features(self):
        return list(self._feature_generator_factories.keys())

    @staticmethod
    def _name(name: Hashable):
        # for enums, which have .name, use the name only, because it is less problematic to persist
        if hasattr(name, "name"):
            name = name.name
        return name

    def register_factory(self, name: Hashable, factory: Callable[[], FeatureGenerator]):
        """
        Registers a feature generator factory which can subsequently be referenced by models via their name/hashable key

        :param name: the name/key (which can, in particular, be a string or an Enum item). Especially for larger projects
            the use of an Enum is recommended (for optimal IDE support)
        :param factory: the factory
        """
        name = self._name(name)
        if name in self._feature_generator_factories:
            raise ValueError(f"Generator for name '{name}' already registered")
        self._feature_generator_factories[name] = factory

    def get_feature_generator(self, name: str) -> FeatureGenerator:
        """
        Creates a feature generator from a name, which must have been previously registered.
        The name of the returned feature generator (as returned by getName()) is set to name.

        :param name: the name (which can, in particular, be a string or an enum item)
        :return: a new feature generator instance (or existing instance for the case where useSingletons is enabled)
        """
        name = self._name(name)
        generator = self._feature_generator_singletons.get(name)
        if generator is None:
            factory = self._feature_generator_factories.get(name)
            if factory is None:
                raise ValueError(f"No factory registered for name '{name}': known names: {list_string(self._feature_generator_factories.keys())}. Use registerFeatureGeneratorFactory to register a new feature generator factory.")
            generator = factory()
            generator.set_name(name)
            if self._use_singletons:
                self._feature_generator_singletons[name] = generator
        return generator

    def collect_features(self, *feature_generators_or_names: Union[Hashable, FeatureGenerator]) -> "FeatureCollector":
        """
        Creates a feature collector for the given feature names/keys/instances, which can subsequently be added to a model.

        :param feature_generators_or_names: feature names/keys known to this registry or feature generator instances
        """
        return FeatureCollector(*feature_generators_or_names, registry=self)


class FeatureCollector(object):
    """
    A feature collector which facilitates the collection of features that shall be used by a model as well as the
    generation of commonly used feature transformers that are informed by the features' meta-data.
    """

    def __init__(self,
            *feature_generators_or_names: Union[Hashable, FeatureGenerator],
            registry: FeatureGeneratorRegistry = None):
        """
        :param feature_generators_or_names: generator names/keys (known to the registry) or generator instances
        :param registry: the feature generator registry for the case where names/keys are passed
        """
        self._feature_generators_or_names = feature_generators_or_names
        self._registry = registry
        self._multi_feature_generator = self.create_multi_feature_generator()

    def get_multi_feature_generator(self) -> MultiFeatureGenerator:
        """
        Gets the multi-feature generator that was created for this collector.
        To create a new, independent instance (e.g. when using this collector for multiple
        models), use :meth:`create_multi_feature_generator` instead.

        :return: the multi-feature generator that was created for this instance
        """
        return self._multi_feature_generator

    def get_normalisation_rules(self, include_generated_categorical_rules=True):
        return self.get_multi_feature_generator().get_normalisation_rules(
            include_generated_categorical_rules=include_generated_categorical_rules)

    def get_categorical_feature_name_regex(self) -> str:
        """
        :return: a regular expression that matches all known categorical feature names
        """
        return self.get_multi_feature_generator().get_categorical_feature_name_regex()

    def create_multi_feature_generator(self):
        """
        Creates a new instance of the multi-feature generator that generates the features
        collected by this instance. If the feature collector instance is not used for
        multiple models, use :meth:`get_multi_feature_generator` instead to obtain
        the instance that has already been created.

        :return: a new multi-feature generator that generates the collected features
        """
        feature_generators = []
        for f in self._feature_generators_or_names:
            if isinstance(f, FeatureGenerator):
                feature_generators.append(f)
            else:
                if self._registry is None:
                    raise Exception(f"Received feature name '{f}' instead of instance but no registry to perform the lookup")
                feature_generators.append(self._registry.get_feature_generator(f))
        return MultiFeatureGenerator(*feature_generators)

    def create_dft_normalisation(self, default_transformer_factory=None, require_all_handled=True, inplace=False) -> DFTNormalisation:
        """
        Creates a feature transformer that will apply normalisation to all supported (numeric) features

        :param default_transformer_factory: a factory for the creation of transformer instances (which implements the
            API used by sklearn.preprocessing, e.g. StandardScaler) that shall be used to create a transformer for all
            rules that do not specify a particular transformer.
            The default transformer will only be applied to columns matched by such rules, unmatched columns will
            not be transformed.
            Use SkLearnTransformerFactoryFactory to conveniently create a factory.
        :param require_all_handled: whether to raise an exception if not all columns are matched by a rule
        :param inplace: whether to apply data frame transformations in-place
        :return: the transformer
        """
        return DFTNormalisation(self.get_normalisation_rules(), default_transformer_factory=default_transformer_factory,
            require_all_handled=require_all_handled, inplace=inplace)

    def create_dft_one_hot_encoder(self, ignore_unknown=False, inplace=False):
        """
        Creates a feature transformer that will apply one-hot encoding to all the features that are known to be categorical

        :param inplace: whether to perform the transformation in-place
        :param ignore_unknown: if True and an unknown category is encountered during transform, the resulting one-hot
            encoded columns for this feature will be all zeros. if False, an unknown category will raise an error.
        :return: the transformer
        """
        return DFTOneHotEncoder(self.get_categorical_feature_name_regex(), ignore_unknown=ignore_unknown, inplace=inplace)

    def create_feature_transformer_normalisation(self, default_transformer_factory=None, require_all_handled=True, inplace=False) \
            -> DFTNormalisation:
        """
        Creates a feature transformer that will apply normalisation to all supported (numeric) features.
        Alias of create_dft_normalisation.

        :param default_transformer_factory: a factory for the creation of transformer instances (which implements the
            API used by sklearn.preprocessing, e.g. StandardScaler) that shall be used to create a transformer for all
            rules that do not specify a particular transformer.
            The default transformer will only be applied to columns matched by such rules, unmatched columns will
            not be transformed.
            Use SkLearnTransformerFactoryFactory to conveniently create a factory.
        :param require_all_handled: whether to raise an exception if not all columns are matched by a rule
        :param inplace: whether to apply data frame transformations in-place
        :return: the transformer
        """
        return self.create_dft_normalisation(default_transformer_factory=default_transformer_factory,
            require_all_handled=require_all_handled, inplace=inplace)

    def create_feature_transformer_one_hot_encoder(self, ignore_unknown=False, inplace=False):
        """
        Creates a feature transformer that will apply one-hot encoding to all the features that are known to be categorical.
        Alias of create_dft_one_hot_encoder.

        :param inplace: whether to perform the transformation in-place
        :param ignore_unknown: if True and an unknown category is encountered during transform, the resulting one-hot
            encoded columns for this feature will be all zeros. if False, an unknown category will raise an error.
        :return: the transformer
        """
        return self.create_dft_one_hot_encoder(ignore_unknown=ignore_unknown, inplace=inplace)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Union, List, Callable

from sensai import VectorRegressionModel, VectorClassificationModel, VectorModelBase
from sensai.evaluation import MultiDataModelEvaluation
from sensai.evaluation.eval_stats import RegressionMetric, ClassificationMetric

TMetric = Union[RegressionMetric, ClassificationMetric]
TModel = Union[VectorClassificationModel, VectorRegressionModel]


@dataclass
class MetricComputationResult:
    metric_value: float
    models: List[VectorModelBase]


class MetricComputation(ABC):
    def __init__(self, metric: TMetric):
        self.metric = metric

    @abstractmethod
    def compute_metric_value(self, model_factory: Callable[[], TModel]) -> MetricComputationResult:
        pass


class MetricComputationMultiData(MetricComputation):
    def __init__(self, ev_util: MultiDataModelEvaluation, use_cross_validation: bool, metric: TMetric,
            use_combined_eval_stats: bool):
        super().__init__(metric)
        self.use_combined_eval_stats = use_combined_eval_stats
        self.ev_util = ev_util
        self.use_cross_validation = use_cross_validation

    def compute_metric_value(self, model_factory: Callable[[], TModel]) -> MetricComputationResult:
        result = self.ev_util.compare_models([model_factory], use_cross_validation=self.use_cross_validation)
        if self.use_combined_eval_stats:
            assert len(result.get_model_names()) == 1, "Model factory must produce named models"
            model_name = result.get_model_names()[0]
            metric_value = result.get_eval_stats_collection(model_name).get_combined_eval_stats().compute_metric_value(self.metric)
            models = []
            for dataset_name, comparison_result in result.iter_model_results(model_name):
                if self.use_cross_validation:
                    models.extend(comparison_result.cross_validation_data.trained_models)
                else:
                    models.append(comparison_result.eval_data.model)
            return MetricComputationResult(metric_value, models)
        else:
            raise NotImplementedError()



from typing import Optional, List, TYPE_CHECKING, Callable

import pandas as pd

from sensai.evaluation.eval_stats import RegressionEvalStats
from sensai.util.pandas import query_data_frame
from sensai.vector_model import get_predicted_var_name

if TYPE_CHECKING:
    from sensai.evaluation import VectorRegressionModelEvaluationData


class ResultSet:
    """
    A result set which is designed for interactive result inspection (e.g. in an iPython notebook).
    An instance can, for example, be created with a data frame as returned by VectorRegressionModelEvaluationData.to_data_frame
    and subsequently be applied to interactively analyse the results.

    The class is designed to be subclassed, such that, in particular, method `_show_df` can be
    overridden to display meaningful information (use case-specific) in the notebook environment.
    """
    def __init__(self, df: pd.DataFrame):
        self.df = df

    def _create_result_set(self, df: pd.DataFrame, parent: "ResultSet"):
        """
        Creates a new result set for the given data frame

        :param df: the data frame
        :return: the result set
        """
        return ResultSet(df)

    def query(self, sql: str) -> "ResultSet":
        """
        Queries the result set with the given condition specified in SQL syntax.

        NOTE: Requires duckdb to be installed.

        :param sql: an SQL query starting with the WHERE clause (excluding the 'where' keyword itself)
        :return: the result set corresponding to the query
        """
        result_df = query_data_frame(self.df, sql)
        return self._create_result_set(result_df, self)

    def show(self, first: Optional[int] = None, sample: Optional[int] = None) -> None:
        """
        Shows all or some of the result set's contents.

        :param first: if not None, show this many rows from the start of the result set
        :param sample: if not None, sample this many rows from the result set to be shown
        """
        df = self.df
        if first is not None:
            df = df.iloc[:first]
        if sample is not None:
            df = df.sample(sample)
        self._show_df(df)

    def _show_df(self, df: pd.DataFrame):
        print(df.to_string())


class RegressionResultSet(ResultSet):
    def __init__(self, df: pd.DataFrame, predicted_var_names: List[str]):
        super().__init__(df)
        self.predicted_var_names = predicted_var_names

    @classmethod
    def from_regression_eval_data(cls, eval_data: "VectorRegressionModelEvaluationData", modify_input_df: bool = False,
            output_col_name_override: Optional[str] = None,
            regression_result_set_factory: Callable[[pd.DataFrame, List[str]], "RegressionResultSet"] = None) \
            -> "RegressionResultSet":
        df = eval_data.to_data_frame(modify_input_df=modify_input_df, output_col_name_override=output_col_name_override)
        if output_col_name_override:
            predicted_var_names = [output_col_name_override]
        else:
            predicted_var_names = eval_data.predicted_var_names

        def default_factory(data_frame: pd.DataFrame, var_names: List[str]):
            return cls(data_frame, var_names)

        if regression_result_set_factory is None:
            regression_result_set_factory = default_factory

        return regression_result_set_factory(df, predicted_var_names)

    def _create_result_set(self, df: pd.DataFrame, parent: "RegressionResultSet"):
        return self.__class__(df, parent.predicted_var_names)

    @staticmethod
    def col_name_predicted(predicted_var_name: str):
        return f"{predicted_var_name}_predicted"

    @staticmethod
    def col_name_ground_truth(predicted_var_name: str):
        return f"{predicted_var_name}_true"

    @staticmethod
    def col_name_error(predicted_var_name: str):
        return f"{predicted_var_name}_error"

    @staticmethod
    def col_name_abs_error(predicted_var_name: str):
        return f"{predicted_var_name}_abs_error"

    def eval_stats(self, predicted_var_name: Optional[str] = None):
        """
        Creates the evaluation stats object for this result object, which can be used to compute metrics
        or to create plots.

        :param predicted_var_name: the name of the predicted variable for which to create the object;
            can be None if there is but a single variable
        :return: the evaluation stats object
        """
        predicted_var_name = get_predicted_var_name(predicted_var_name, self.predicted_var_names)
        return RegressionEvalStats(y_predicted=self.df[self.col_name_predicted(predicted_var_name)],
            y_true=self.df[self.col_name_ground_truth(predicted_var_name)])



import enum
import logging
import pickle
import pandas as pd

from .cache import PersistentKeyValueCache, DelayedUpdateHook

log = logging.getLogger(__name__)


class MySQLPersistentKeyValueCache(PersistentKeyValueCache):

    class ValueType(enum.Enum):
        DOUBLE = ("DOUBLE", False)  # (SQL data type, isCachedValuePickled)
        BLOB = ("BLOB", True)

    def __init__(self, host, db, user, pw, value_type: ValueType, table_name="cache", deferred_commit_delay_secs=1.0, in_memory=False):
        import MySQLdb
        self.conn = MySQLdb.connect(host=host, database=db, user=user, password=pw)
        self.table_name = table_name
        self.max_key_length = 255
        self._update_hook = DelayedUpdateHook(self._commit, deferred_commit_delay_secs)
        self._num_entries_to_be_committed = 0

        cache_value_sql_type, self.is_cache_value_pickled = value_type.value

        cursor = self.conn.cursor()
        cursor.execute(f"SHOW TABLES;")
        if table_name not in [r[0] for r in cursor.fetchall()]:
            cursor.execute(f"CREATE TABLE {table_name} (cache_key VARCHAR({self.max_key_length}) PRIMARY KEY, "
                           f"cache_value {cache_value_sql_type});")
        cursor.close()

        self._in_memory_df = None if not in_memory else self._load_table_to_data_frame()

    def _load_table_to_data_frame(self):
        df = pd.read_sql(f"SELECT * FROM {self.table_name};", con=self.conn, index_col="cache_key")
        if self.is_cache_value_pickled:
            df["cache_value"] = df["cache_value"].apply(pickle.loads)
        return df

    def set(self, key, value):
        key = str(key)
        if len(key) > self.max_key_length:
            raise ValueError(f"Key too long, maximal key length is {self.max_key_length}")
        cursor = self.conn.cursor()
        cursor.execute(f"SELECT COUNT(*) FROM {self.table_name} WHERE cache_key=%s", (key,))
        stored_value = pickle.dumps(value) if self.is_cache_value_pickled else value
        if cursor.fetchone()[0] == 0:
            cursor.execute(f"INSERT INTO {self.table_name} (cache_key, cache_value) VALUES (%s, %s)",
                (key, stored_value))
        else:
            cursor.execute(f"UPDATE {self.table_name} SET cache_value=%s WHERE cache_key=%s", (stored_value, key))
        self._num_entries_to_be_committed += 1
        self._update_hook.handle_update()
        cursor.close()
        if self._in_memory_df is not None:
            self._in_memory_df["cache_value"][str(key)] = value

    def get(self, key):
        value = self._get_from_in_memory_df(key)
        if value is None:
            value = self._get_from_table(key)
        return value

    def _get_from_table(self, key):
        cursor = self.conn.cursor()
        cursor.execute(f"SELECT cache_value FROM {self.table_name} WHERE cache_key=%s", (str(key),))
        row = cursor.fetchone()
        if row is None:
            return None
        stored_value = row[0]
        value = pickle.loads(stored_value) if self.is_cache_value_pickled else stored_value
        return value

    def _get_from_in_memory_df(self, key):
        if self._in_memory_df is None:
            return None
        try:
            return self._in_memory_df["cache_value"][str(key)]
        except Exception as e:
            log.debug(f"Unable to load value for key {str(key)} from in-memory dataframe: {e}")
            return None

    def _commit(self):
        log.info(f"Committing {self._num_entries_to_be_committed} cache entries to the database")
        self.conn.commit()
        self._num_entries_to_be_committed = 0



"""
This module contains various helper functions.
"""
import math
from typing import Any, Sequence, Union, TypeVar, List, Optional, Dict, Container, Iterable

T = TypeVar("T")


def count_none(*args: Any) -> int:
    """
    Counts the number of arguments that are None

    :param args: various arguments
    :return: the number of arguments that are None
    """
    c = 0
    for a in args:
        if a is None:
            c += 1
    return c


def count_not_none(*args: Any) -> int:
    """
    Counts the number of arguments that are not None

    :param args: various arguments
    :return: the number of arguments that are not None
    """
    return len(args) - count_none(*args)


def any_none(*args: Any) -> bool:
    """
    :param args: various arguments
    :return: True if any of the arguments are None, False otherwise
    """
    return count_none(*args) > 0


def all_none(*args: Any) -> bool:
    """
    :param args: various arguments
    :return: True if all of the arguments are None, False otherwise
    """
    return count_none(*args) == len(args)


def check_not_nan_dict(d: dict):
    """
    Raises ValueError if any of the values in the given dictionary are NaN, reporting the respective keys

    :param d: a dictionary mapping to floats that are to be checked for NaN
    """
    invalid_keys = [k for k, v in d.items() if math.isnan(v)]
    if len(invalid_keys) > 0:
        raise ValueError(f"Got one or more NaN values: {invalid_keys}")


def contains_any(container: Union[Container, Iterable], items: Sequence) -> bool:
    for item in items:
        if item in container:
            return True
    return False


# noinspection PyUnusedLocal
def mark_used(*args):
    """
    Utility function to mark identifiers as used.
    The function does nothing.

    :param args: pass identifiers that shall be marked as used here
    """
    pass


def flatten_arguments(args: Sequence[Union[T, Sequence[T]]]) -> List[T]:
    """
    Main use case is to support both interfaces of the type f(T1, T2, ...) and f([T1, T2, ...]) simultaneously.
    It is assumed that if the latter form is passed, the arguments are either in a list or a tuple. Moreover,
    T cannot be a tuple or a list itself.

    Overall this function is not all too safe and one should be aware of what one is doing when using it
    """
    result = []
    for arg in args:
        if isinstance(arg, list) or isinstance(arg, tuple):
            result.extend(arg)
        else:
            result.append(arg)
    return result


def kwarg_if_not_none(arg_name: str, arg_value: Any) -> Dict[str, Any]:
    """
    Supports the optional passing of a kwarg, returning a non-empty dictionary with the kwarg only
    if the value is not None.

    This can be helpful to improve backward compatibility for cases where a kwarg was added later
    but old implementations weren't updated to have it, such that an exception will be raised only
    if the kwarg is actually used.

    :param arg_name: the argument name
    :param arg_value: the value
    :return: a dictionary containing the kwarg or, if the value is None, an empty dictionary
    """
    if arg_value is None:
        return {}
    else:
        return {arg_name: arg_value}



import copy
import logging
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Any, Dict, Optional

import numpy as np
import pandas as pd
from sklearn import compose

from ..feature_importance import FeatureImportanceProvider
from ..util.pickle import setstate
from ..util.string import dict_string
from ..vector_model import VectorRegressionModel, VectorClassificationModel

log = logging.getLogger(__name__)


def create_sklearn_model(model_constructor, model_args, output_transformer=None):
    model = model_constructor(**model_args)
    if output_transformer is not None:
        model = compose.TransformedTargetRegressor(regressor=model, transformer=output_transformer)
    return model


def str_sklearn_model(model):
    """
    Creates a cleaned string representation of the model with line breaks and indentations removed

    :param model: the sklearn model for which to generate the cleaned string representation
    :return: the string representation
    """
    return re.sub(r",\s*", ", ", str(model))


def _apply_sklearn_input_transformer(inputs: pd.DataFrame, sklearn_input_transformer: Optional, fit: bool) -> pd.DataFrame:
    if sklearn_input_transformer is None:
        return inputs
    else:
        input_values = inputs.values
        shape_before = input_values.shape
        if fit:
            input_values = sklearn_input_transformer.fit_transform(input_values)
        else:
            input_values = sklearn_input_transformer.transform(input_values)
        if input_values.shape != shape_before:
            raise Exception("sklearnInputTransformer changed the shape of the input, which is unsupported. "
                            "Consider using an a DFTSkLearnTransformer as a feature transformer instead.")
        return pd.DataFrame(input_values, index=inputs.index, columns=inputs.columns)


class ActualFitParams:
    def __init__(self, inputs, outputs, kwargs: Dict[str, Any]):
        self.inputs = inputs
        self.outputs = outputs
        self.kwargs = kwargs


class AbstractSkLearnVectorRegressionModel(VectorRegressionModel, ABC):
    """
    Base class for models built upon scikit-learn's model implementations
    """
    log = log.getChild(__qualname__)

    def __init__(self, model_constructor, **model_args):
        """
        :param model_constructor: the sklearn model constructor
        :param model_args: arguments to be passed to the sklearn model constructor
        """
        super().__init__()
        self.sklearnInputTransformer = None
        self.sklearnOutputTransformer = None
        self.modelConstructor = model_constructor
        self.modelArgs = model_args
        self.fitArgs = {}

    def _tostring_excludes(self) -> List[str]:
        return super()._tostring_excludes() + ["sklearnInputTransformer", "sklearnOutputTransformer", "modelConstructor", "modelArgs"]

    def with_sklearn_input_transformer(self, sklearn_input_transformer) -> __qualname__:
        """
        :param sklearn_input_transformer: an optional sklearn preprocessor for normalising/scaling inputs
        :return: self
        """
        self.sklearnInputTransformer = sklearn_input_transformer
        return self

    def with_sklearn_output_transformer(self, sklearn_output_transformer):
        """
        :param sklearn_output_transformer: an optional sklearn preprocessor for normalising/scaling outputs
        :return: self
        """
        self.sklearnOutputTransformer = sklearn_output_transformer
        return self

    def _transform_input(self, inputs: pd.DataFrame, fit=False) -> pd.DataFrame:
        return _apply_sklearn_input_transformer(inputs, self.sklearnInputTransformer, fit)

    def _update_model_args(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        """
        Designed to be overridden in order to make input data-specific changes to modelArgs

        :param inputs: the training input data
        :param outputs: the training output data
        """
        pass

    def _update_fit_args(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        """
        Designed to be overridden in order to make input data-specific changes to fitArgs (arguments to be passed to the
        underlying model's `fit` method)

        :param inputs: the training input data
        :param outputs: the training output data
        """
        pass

    def _compute_actual_fit_params(self, inputs: pd.DataFrame, outputs: pd.DataFrame, weights: Optional[pd.Series] = None) -> ActualFitParams:
        """
        Computes additional arguments to be passed to the model's `fit` method, which are transient and shall not be saved
        along with the model as metadata, e.g. larger data structures such as validation data or sample weights.

        :param inputs: the training input data
        :param outputs: the training output data
        :return: a dictionary of parameters to be passed to `fit`.
        """
        fit_params = ActualFitParams(inputs, outputs, dict(self.fitArgs))
        if weights is not None:
            self._warn_sample_weights_unsupported(self.is_sample_weight_supported(), weights)
            if self.is_sample_weight_supported():
                fit_params.kwargs["sample_weight"] = weights
        return fit_params

    @abstractmethod
    def is_sample_weight_supported(self) -> bool:
        pass

    def _fit(self, inputs: pd.DataFrame, outputs: pd.DataFrame, weights: Optional[pd.Series] = None):
        inputs = self._transform_input(inputs, fit=True)
        self._update_model_args(inputs, outputs)
        self._update_fit_args(inputs, outputs)
        actual_fit_params = self._compute_actual_fit_params(inputs, outputs, weights=weights)
        self._fit_sklearn(actual_fit_params)

    @abstractmethod
    def _fit_sklearn(self, params: ActualFitParams):
        pass

    def _predict(self, x: pd.DataFrame):
        inputs = self._transform_input(x)
        return self._predict_sklearn(inputs)

    @abstractmethod
    def _predict_sklearn(self, inputs: pd.DataFrame):
        pass


class AbstractSkLearnMultipleOneDimVectorRegressionModel(AbstractSkLearnVectorRegressionModel, ABC):
    """
    Base class for models which use several sklearn models of the same type with a single
    output dimension to create a multi-dimensional model (for the case where there is more than one output dimension)
    """
    def __init__(self, model_constructor, **model_args):
        super().__init__(model_constructor, **model_args)
        self.models = {}

    def _tostring_excludes(self) -> List[str]:
        return super()._tostring_excludes() + ["models"]

    def _tostring_additional_entries(self) -> Dict[str, Any]:
        d = super()._tostring_additional_entries()
        if len(self.models) > 0:
            d["model[0]"] = str_sklearn_model(next(iter(self.models.values())))
        else:
            d["modelConstructor"] = f"{self.modelConstructor.__name__}({dict_string(self.modelArgs)})"
        return d

    def _fit_sklearn(self, params: ActualFitParams):
        for predictedVarName in params.outputs.columns:
            log.info(f"Fitting model for output variable '{predictedVarName}'")
            model = create_sklearn_model(self.modelConstructor,
                    self.modelArgs,
                    output_transformer=copy.deepcopy(self.sklearnOutputTransformer))
            model.fit(params.inputs, params.outputs[predictedVarName], **params.kwargs)
            self.models[predictedVarName] = model

    def _predict_sklearn(self, inputs: pd.DataFrame) -> pd.DataFrame:
        results = {}
        for varName in self.models:
            results[varName] = self._predict_sklearn_single_model(self.models[varName], inputs)
        return pd.DataFrame(results)

    def _predict_sklearn_single_model(self, model, inputs: pd.DataFrame) -> np.ndarray:
        return model.predict(inputs)

    def get_sklearn_model(self, predicted_var_name=None):
        if predicted_var_name is None:
            if len(self.models) > 1:
                raise ValueError(f"Must provide predicted variable name (one of {self.models.keys()})")
            return next(iter(self.models.values()))
        return self.models[predicted_var_name]


class AbstractSkLearnMultiDimVectorRegressionModel(AbstractSkLearnVectorRegressionModel, ABC):
    """
    Base class for models which use a single sklearn model with multiple output dimensions to create the multi-dimensional model
    """
    def __init__(self, model_constructor, **model_args):
        super().__init__(model_constructor, **model_args)
        self.model = None

    def _tostring_excludes(self) -> List[str]:
        return super()._tostring_excludes() + ["model"]

    def _tostring_additional_entries(self) -> Dict[str, Any]:
        d = super()._tostring_additional_entries()
        if self.model is not None:
            d["model"] = str_sklearn_model(self.model)
        else:
            d["modelConstructor"] = f"{self.modelConstructor.__name__}({dict_string(self.modelArgs)})"
        return d

    def _fit_sklearn(self, params: ActualFitParams):
        if len(params.outputs.columns) > 1:
            log.info(f"Fitting a single multi-dimensional model for all {len(params.outputs.columns)} output dimensions")
        self.model = create_sklearn_model(self.modelConstructor, self.modelArgs, output_transformer=self.sklearnOutputTransformer)
        output_values = params.outputs.values
        if output_values.shape[1] == 1:  # for 1D output, shape must be (numSamples,) rather than (numSamples, 1)
            output_values = np.ravel(output_values)
        self.model.fit(params.inputs, output_values, **params.kwargs)

    def _predict_sklearn(self, inputs: pd.DataFrame) -> pd.DataFrame:
        y = self.model.predict(inputs)
        return pd.DataFrame(y, columns=self.get_model_output_variable_names())


class AbstractSkLearnVectorClassificationModel(VectorClassificationModel, ABC):
    def __init__(self, model_constructor, use_balanced_class_weights=False, use_label_encoding=False, **model_args):
        """
        :param model_constructor: the sklearn model constructor
        :param model_args: arguments to be passed to the sklearn model constructor
        :param use_balanced_class_weights: whether to compute class weights from the training data and apply the corresponding weight to
            each data point such that the sum of weights for all classes is equal. This is achieved by applying a weight proportional
            to the reciprocal frequency of the class in the (training) data. We scale weights such that the smallest weight (of the
            largest class) is 1, ensuring that weight counts still reasonably correspond to data point counts.
            Note that weighted data points may not be supported for all types of models.
        :param use_label_encoding: whether to replace original class labels with 0-based index in sorted list of labels (a.k.a. label
            encoding), which is required by some sklearn-compatible implementations (particularly xgboost)
        """
        super().__init__()
        self.modelConstructor = model_constructor
        self.sklearnInputTransformer = None
        self.modelArgs = model_args
        self.fitArgs = {}
        self.useBalancedClassWeights = use_balanced_class_weights
        self.useLabelEncoding = use_label_encoding
        self.model = None

    def __setstate__(self, state):
        setstate(AbstractSkLearnVectorClassificationModel, self, state, new_optional_properties=["labelEncoder"],
            new_default_properties={"useComputedClassWeights": False, "useLabelEncoder": False},
            renamed_properties={"useComputedClassWeights": "useBalancedClassWeights"},
            removed_properties=["sklearnOutputTransformer"])

    def _tostring_excludes(self) -> List[str]:
        return super()._tostring_excludes() + ["modelConstructor", "sklearnInputTransformer", "sklearnOutputTransformer",
            "modelArgs", "model"]

    def _tostring_additional_entries(self) -> Dict[str, Any]:
        d = super()._tostring_additional_entries()
        if self.model is None:
            d["modelConstructor"] = f"{self.modelConstructor.__name__}({dict_string(self.modelArgs)})"
        else:
            d["model"] = str_sklearn_model(self.model)
        return d

    def with_sklearn_input_transformer(self, sklearn_input_transformer) -> __qualname__:
        """
        :param sklearn_input_transformer: an optional sklearn preprocessor for transforming inputs
        :return: self
        """
        self.sklearnInputTransformer = sklearn_input_transformer
        return self

    def _update_model_args(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        """
        Designed to be overridden in order to make input data-specific changes to modelArgs

        :param inputs: the training input data
        :param outputs: the training output data
        """
        pass

    def _update_fit_args(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        """
        Designed to be overridden in order to make input data-specific changes to fitArgs (arguments to be passed to the
        underlying model's fit method)

        :param inputs: the training input data
        :param outputs: the training output data
        """
        pass

    @abstractmethod
    def is_sample_weight_supported(self) -> bool:
        pass

    def _fit_classifier(self, inputs: pd.DataFrame, outputs: pd.DataFrame, weights: Optional[pd.Series] = None):
        inputs = self._transform_input(inputs, fit=True)
        self._update_model_args(inputs, outputs)
        self._update_fit_args(inputs, outputs)
        self.model = create_sklearn_model(self.modelConstructor, self.modelArgs)
        log.info(f"Fitting sklearn classifier of type {self.model.__class__.__name__}")
        kwargs = dict(self.fitArgs)

        if self.useBalancedClassWeights and weights is not None:
            raise ValueError("Balanced class weights cannot be used in conjunction with user-specified weights")

        if self.useBalancedClassWeights:
            class2weight = self._compute_class_weights(outputs)
            classes = outputs.iloc[:, 0]
            weights = np.array([class2weight[cls] for cls in classes])
            weights = weights / np.min(weights)
            kwargs["sample_weight"] = weights

        elif weights is not None:
            self._warn_sample_weights_unsupported(self.is_sample_weight_supported(), weights)
            if self.is_sample_weight_supported():
                kwargs["sample_weight"] = weights

        output_values = np.ravel(outputs.values)
        if self.useLabelEncoding:
            output_values = self._encode_labels(output_values)
        self._fit_sklearn_classifier(inputs, output_values, kwargs)

    def _fit_sklearn_classifier(self, inputs: pd.DataFrame, output_values: np.ndarray, kwargs: Dict[str, Any]):
        self.model.fit(inputs, output_values, **kwargs)

    def _transform_input(self, inputs: pd.DataFrame, fit=False) -> pd.DataFrame:
        return _apply_sklearn_input_transformer(inputs, self.sklearnInputTransformer, fit)

    def _encode_labels(self, y: np.ndarray):
        d = {l: i for i, l in enumerate(self._labels)}
        vfn = np.vectorize(lambda x: d[x])
        return vfn(y)

    def _decode_labels(self, y: np.ndarray):
        d = dict(enumerate(self._labels))
        vfn = np.vectorize(lambda x: d[x])
        return vfn(y)

    def _predict_sklearn(self, input_values):
        return self.model.predict(input_values)

    def _predict(self, x: pd.DataFrame):
        input_values = self._transform_input(x)
        y = self._predict_sklearn(input_values)
        if self.useLabelEncoding:
            y = self._decode_labels(y)
        return pd.DataFrame(y, columns=self._predictedVariableNames)

    def _predict_class_probabilities(self, x: pd.DataFrame):
        input_values = self._transform_input(x)
        y = self.model.predict_proba(input_values)
        return pd.DataFrame(y, columns=self._labels)

    def get_params(self, deep=True):
        return self.model.get_params(deep=deep)

    def set_params(self, **params):
        self.model.set_params(**params)

    # noinspection PyMethodMayBeStatic
    def _compute_class_weights(self, outputs: pd.DataFrame):
        """
        :param outputs: the output data frame containing the class labels as the first column
        :return: the dictionary of class weights mapping class to weight value
        """
        classes: pd.Series = outputs.iloc[:, 0]
        counts = classes.value_counts()
        rfreqs = counts / counts.sum()
        weights: pd.Series = 1.0 / rfreqs
        return weights.to_dict()


def _get_model_feature_importance_vector(model):
    cand_attributes = ("feature_importances_", "coef_")
    for attr in cand_attributes:
        if hasattr(model, attr):
            importance_values = getattr(model, attr)
            if attr == "coef_":
                importance_values = np.abs(importance_values)  # for coefficients in linear models, use the absolute values
            return importance_values
    raise ValueError(f"Model {model} has none of the attributes {cand_attributes}")


class FeatureImportanceProviderSkLearnRegressionMultipleOneDim(FeatureImportanceProvider):
    def get_feature_importance_dict(self) -> Dict[str, Dict[str, int]]:
        self: AbstractSkLearnMultipleOneDimVectorRegressionModel
        return {targetFeature: dict(zip(self._modelInputVariableNames, _get_model_feature_importance_vector(model)))
                for targetFeature, model in self.models.items()}


class FeatureImportanceProviderSkLearnRegressionMultiDim(FeatureImportanceProvider):
    def get_feature_importance_dict(self) -> Dict[str, float]:
        self: AbstractSkLearnMultiDimVectorRegressionModel
        return dict(zip(self._modelInputVariableNames, _get_model_feature_importance_vector(self.model)))


class FeatureImportanceProviderSkLearnClassification(FeatureImportanceProvider):
    def get_feature_importance_dict(self) -> Dict[str, float]:
        self: AbstractSkLearnVectorClassificationModel
        return dict(zip(self._modelInputVariableNames, _get_model_feature_importance_vector(self.model)))



"""
This module exists only for backward compatibility with object that were pickled by earlier versions of sensAI.
"""

from .torch_base import MCDropoutCapableNNModule
from .torch_models.mlp.mlp_modules import MultiLayerPerceptron
from .torch_models.lstnet.lstnet_modules import LSTNetwork
from .torch_models.residualffn.residualffn_modules import ResidualFeedForwardNetwork
from ..util import mark_used

mark_used(MCDropoutCapableNNModule, MultiLayerPerceptron, LSTNetwork, ResidualFeedForwardNetwork)


import atexit
import logging as lg
import sys
import time
from abc import ABC, abstractmethod
from datetime import datetime
from io import StringIO
from logging import *
from typing import List, Callable, Optional, TypeVar, TYPE_CHECKING, Generic

from .time import format_duration

if TYPE_CHECKING:
    import pandas as pd


log = getLogger(__name__)

LOG_DEFAULT_FORMAT = '%(levelname)-5s %(asctime)-15s %(name)s:%(funcName)s:%(lineno)d - %(message)s'
T = TypeVar("T")
THandler = TypeVar("THandler", bound=Handler)

# Holds the log format that is configured by the user (using function `configure`), such
# that it can be reused in other places
_logFormat = LOG_DEFAULT_FORMAT

# User-configured callback which is called after logging is configured via function `configure`
_configureCallbacks: List[Callable[[], None]] = []


def remove_log_handlers():
    """
    Removes all current log handlers
    """
    logger = getLogger()
    while logger.hasHandlers():
        logger.removeHandler(logger.handlers[0])


def remove_log_handler(handler):
    getLogger().removeHandler(handler)


def is_log_handler_active(handler):
    return handler in getLogger().handlers


def is_enabled() -> bool:
    """
    :return: True if logging is enabled (at least one log handler exists)
    """
    return getLogger().hasHandlers()


def set_configure_callback(callback: Callable[[], None], append: bool = True) -> None:
    """
    Configures a function to be called when logging is configured, e.g. through :func:`configure, :func:`run_main` or
    :func:`run_cli`.
    A typical use for the callback is to configure the logging behaviour of packages, setting appropriate log levels.

    :param callback: the function to call
    :param append: whether to append to the list of callbacks; if False, any existing callbacks will be removed
    """
    global _configureCallbacks
    if not append:
        _configureCallbacks = []
    _configureCallbacks.append(callback)


# noinspection PyShadowingBuiltins
def configure(format=LOG_DEFAULT_FORMAT, level=lg.DEBUG):
    """
    Configures logging to stdout with the given format and log level,
    also configuring the default log levels of some overly verbose libraries as well as some pandas output options.

    :param format: the log format
    :param level: the minimum log level
    """
    global _logFormat
    _logFormat = format
    remove_log_handlers()
    basicConfig(level=level, format=format, stream=sys.stdout)
    getLogger("matplotlib").setLevel(lg.INFO)
    getLogger("urllib3").setLevel(lg.INFO)
    getLogger("msal").setLevel(lg.INFO)
    try:
        import pandas as pd
        pd.set_option('display.max_colwidth', 255)
    except ImportError:
        pass
    for callback in _configureCallbacks:
        callback()


# noinspection PyShadowingBuiltins
def run_main(main_fn: Callable[..., T], format=LOG_DEFAULT_FORMAT, level=lg.DEBUG) -> T:
    """
    Configures logging with the given parameters, ensuring that any exceptions that occur during
    the execution of the given function are logged.
    Logs two additional messages, one before the execution of the function, and one upon its completion.

    :param main_fn: the function to be executed
    :param format: the log message format
    :param level: the minimum log level
    :return: the result of `main_fn`
    """
    configure(format=format, level=level)
    log.info("Starting")
    try:
        result = main_fn()
        log.info("Done")
        return result
    except Exception as e:
        log.error("Exception during script execution", exc_info=e)


# noinspection PyShadowingBuiltins
def run_cli(main_fn: Callable[..., T], format: str = LOG_DEFAULT_FORMAT, level: int = lg.DEBUG) -> Optional[T]:
    """
    Configures logging with the given parameters and runs the given main function as a
    CLI using `jsonargparse` (which is configured to also parse attribute docstrings, such
    that dataclasses can be used as function arguments).
    Using this function requires that `jsonargparse` and `docstring_parser` be available.
    Like `run_main`, two additional log messages will be logged (at the beginning and end
    of the execution), and it is ensured that all exceptions will be logged.

    :param main_fn: the function to be executed
    :param format: the log message format
    :param level: the minimum log level
    :return: the result of `main_fn`
    """
    from jsonargparse import set_docstring_parse_options, CLI

    set_docstring_parse_options(attribute_docstrings=True)
    return run_main(lambda: CLI(main_fn), format=format, level=level)


def datetime_tag() -> str:
    """
    :return: a string tag for use in log file names which contains the current date and time (compact but readable)
    """
    return datetime.now().strftime('%Y%m%d-%H%M%S')


_fileLoggerPaths: List[str] = []
_isAtExitReportFileLoggerRegistered = False


def _at_exit_report_file_logger():
    for path in _fileLoggerPaths:
        print(f"A log file was saved to {path}")


def add_file_logger(path, append=True, register_atexit=True) -> FileHandler:
    global _isAtExitReportFileLoggerRegistered
    log.info(f"Logging to {path} ...")
    mode = "a" if append else "w"
    handler = FileHandler(path, mode=mode)
    handler.setFormatter(Formatter(_logFormat))
    Logger.root.addHandler(handler)
    _fileLoggerPaths.append(path)
    if not _isAtExitReportFileLoggerRegistered and register_atexit:
        atexit.register(_at_exit_report_file_logger)
        _isAtExitReportFileLoggerRegistered = True
    return handler


class MemoryStreamHandler(StreamHandler):
    def __init__(self, stream: StringIO):
        super().__init__(stream)

    def get_log(self) -> str:
        stream: StringIO = self.stream
        return stream.getvalue()


def add_memory_logger() -> MemoryStreamHandler:
    """
    Adds an in-memory logger, i.e. all log statements are written to a memory buffer which can be retrieved
    using the handler's `get_log` method.
    """
    handler = MemoryStreamHandler(StringIO())
    handler.setFormatter(Formatter(_logFormat))
    Logger.root.addHandler(handler)
    return handler


class StopWatch:
    """
    Represents a stop watch for timing an execution. Constructing an instance starts the stopwatch.
    """
    def __init__(self, start=True):
        self.start_time = time.time()
        self._elapsed_secs = 0.0
        self.is_running = start

    def reset(self, start=True):
        """
        Resets the stopwatch, setting the elapsed time to zero.

        :param start: whether to start the stopwatch immediately
        """
        self.start_time = time.time()
        self._elapsed_secs = 0.0
        self.is_running = start

    def restart(self):
        """
        Resets the stopwatch (setting the elapsed time to zero) and restarts it immediately
        """
        self.reset(start=True)

    def _get_elapsed_time_since_last_start(self):
        if self.is_running:
            return time.time() - self.start_time
        else:
            return 0

    def pause(self):
        """
        Pauses the stopwatch. It can be resumed via method 'resume'.
        """
        self._elapsed_secs += self._get_elapsed_time_since_last_start()
        self.is_running = False

    def resume(self):
        """
        Resumes the stopwatch (assuming that it is currently paused). If the stopwatch is not paused,
        the method has no effect (and a warning is logged).
        """
        if not self.is_running:
            self.start_time = time.time()
            self.is_running = True
        else:
            log.warning("Stopwatch is already running (resume has not effect)")

    def get_elapsed_time_secs(self) -> float:
        """
        Gets the total elapsed time, in seconds, on this stopwatch.

        :return: the elapsed time in seconds
        """
        return self._elapsed_secs + self._get_elapsed_time_since_last_start()

    def get_elapsed_timedelta(self) -> "pd.Timedelta":
        """
        :return: the elapsed time as a pandas.Timedelta object
        """
        return pd.Timedelta(self.get_elapsed_time_secs(), unit="s")

    def get_elapsed_time_string(self) -> str:
        """
        :return: a string representation of the elapsed time
        """
        secs = self.get_elapsed_time_secs()
        if secs < 60:
            return f"{secs:.3f} seconds"
        else:
            try:
                import pandas as pd
                return str(pd.Timedelta(secs, unit="s"))
            except ImportError:
                return format_duration(secs)


class StopWatchManager:
    """
    A singleton which manages a pool of named stopwatches, such that executions to be timed by referring to a name only -
    without the need for a limited scope.
    """
    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = StopWatchManager(42)
        return cls._instance

    def __init__(self, secret):
        if secret != 42:
            raise Exception("Use only the singleton instance via getInstance")
        self._stopWatches = {}

    def start(self, name):
        self._stopWatches[name] = time.time()

    def stop(self, name) -> float:
        """
        :param name: the name of the stopwatch
        :return: the time that has passed in seconds
        """
        time_passed_secs = time.time() - self._stopWatches[name]
        del self._stopWatches[name]
        return time_passed_secs

    def is_running(self, name):
        return name in self._stopWatches


class LogTime:
    """
    An execution time logger which can be conveniently applied using a with-statement - in order to log the executing time of the respective
    with-block.
    """

    def __init__(self, name, enabled=True, logger: Logger = None):
        """
        :param name: the name of the event whose time is to be logged upon completion as "<name> completed in <time>"
        :param enabled: whether the logging is actually enabled; can be set to False to disable logging without necessitating
            changes to client code
        :param logger: the logger to use; if None, use the logger of LogTime's module
        """
        self.name = name
        self.enabled = enabled
        self.stopwatch = None
        self.logger = logger if logger is not None else log

    def start(self):
        """
        Starts the stopwatch
        """
        self.stopwatch = StopWatch()
        if self.enabled:
            self.logger.info(f"{self.name} starting ...")

    def stop(self):
        """
        Stops the stopwatch and logs the time taken (if enabled)
        """
        if self.stopwatch is not None and self.enabled:
            self.logger.info(f"{self.name} completed in {self.stopwatch.get_elapsed_time_string()}")

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def __enter__(self):
        self.start()
        return self


class LoggerContext(Generic[THandler], ABC):
    """
    Base class for context handlers to be used in conjunction with Python's `with` statement.
    """

    def __init__(self, enabled=True):
        """
        :param enabled: whether to actually perform any logging.
            This switch allows the with statement to be applied regardless of whether logging shall be enabled.
        """
        self.enabled = enabled
        self._log_handler = None

    @abstractmethod
    def _create_log_handler(self) -> THandler:
        pass

    def __enter__(self) -> Optional[THandler]:
        if self.enabled:
            self._log_handler = self._create_log_handler()
        return self._log_handler

    def __exit__(self, exc_type, exc_value, traceback):
        if self._log_handler is not None:
            remove_log_handler(self._log_handler)


class FileLoggerContext(LoggerContext[FileHandler]):
    """
    A context handler to be used in conjunction with Python's `with` statement which enables file-based logging.
    """

    def __init__(self, path: str, append=True, enabled=True):
        """
        :param path: the path to the log file
        :param append: whether to append in case the file already exists; if False, always create a new file.
        :param enabled: whether to actually perform any logging.
            This switch allows the with statement to be applied regardless of whether logging shall be enabled.
        """
        self.path = path
        self.append = append
        super().__init__(enabled=enabled)

    def _create_log_handler(self) -> FileHandler:
        return add_file_logger(self.path, append=self.append, register_atexit=False)


class MemoryLoggerContext(LoggerContext[MemoryStreamHandler]):
    """
    A context handler to be used in conjunction with Python's `with` statement which enables in-memory logging.
    """

    def _create_log_handler(self) -> MemoryStreamHandler:
        return add_memory_logger()


class LoggingDisabledContext:
    """
    A context manager that will temporarily disable logging
    """
    def __init__(self, highest_level=CRITICAL):
        """
        :param highest_level: the highest level to disable
        """
        self._highest_level = highest_level
        self._previous_level = None

    def __enter__(self):
        self._previous_level = lg.root.manager.disable
        lg.disable(self._highest_level)

    def __exit__(self, exc_type, exc_value, traceback):
        lg.disable(self._previous_level)


# noinspection PyShadowingBuiltins
class FallbackHandler(Handler):
    """
    A log handler which applies the given handler only if the root logger has no defined handlers (or no handlers other than
    this fallback handler itself)
    """
    def __init__(self, handler: Handler, level=NOTSET):
        """
        :param handler: the handler to which messages shall be passed on
        :param level: the minimum level to output; if NOTSET, pass on everything
        """
        super().__init__(level=level)
        self._handler = handler

    def emit(self, record):
        root_handlers = getLogger().handlers
        if len(root_handlers) == 0 or (len(root_handlers) == 1 and root_handlers[0] == self):
            self._handler.emit(record)



import logging
from copy import copy
from dataclasses import dataclass
from typing import Union, List, Callable

import matplotlib.pyplot as plt
import numpy as np

from sensai import VectorModel, InputOutputData, VectorClassificationModel, VectorRegressionModel
from sensai.data_transformation import DFTColumnFilter
from sensai.evaluation import VectorModelCrossValidatorParams, create_vector_model_cross_validator
from sensai.evaluation.metric_computation import MetricComputation
from sensai.feature_importance import FeatureImportanceProvider, AggregatedFeatureImportance
from sensai.util.plot import ScatterPlot

log = logging.getLogger(__name__)


@dataclass
class RFEStep:
    metric_value: float
    features: List[str]


class RFEResult:
    def __init__(self, steps: List[RFEStep], metric_name: str, minimise: bool):
        self.steps = steps
        self.metric_name = metric_name
        self.minimise = minimise

    def get_sorted_steps(self) -> List[RFEStep]:
        """
        :return: the elimination step results, sorted from best to worst
        """
        return sorted(self.steps, key=lambda s: s.metric_value, reverse=not self.minimise)

    def get_selected_features(self) -> List[str]:
        return self.get_sorted_steps()[0].features

    def get_num_features_array(self) -> np.ndarray:
        """
        :return: array containing the number of features that was considered in each step
        """
        return np.array([len(s.features) for s in self.steps])

    def get_metric_values_array(self) -> np.ndarray:
        """
        :return: array containing the metric value that resulted in each step
        """
        return np.array([s.metric_value for s in self.steps])

    def plot_metric_values(self) -> plt.Figure:
        """
        Plots the metric values vs. the number of features for each step of the elimination

        :return: the figure
        """
        return ScatterPlot(self.get_num_features_array(), self.get_metric_values_array(), c_opacity=1, x_label="number of features",
            y_label=f"cross-validation mean metric value ({self.metric_name})").fig


class RecursiveFeatureEliminationCV:
    """
    Recursive feature elimination, using cross-validation to select the best set of features:
    In each step, the model is first evaluated using cross-validation.
    Then the feature importance values are aggregated across the models that were trained during cross-validation,
    and the least important feature is discarded. For the case where the lowest feature importance is 0, all
    features with 0 importance are discarded.
    This process is repeated until a point is reached where only `minFeatures` (or less) remain.
    The selected set of features is the one from the step where cross-validation yielded the best evaluation metric value.

    Feature importance is computed at the level of model input features, i.e. after feature generation and transformation.

    NOTE: This implementation differs markedly from sklearn's RFECV, which performs an independent RFE for each fold.
    RFECV determines the number of features to use by determining the elimination step in each fold that yielded the best
    metric value on average. Because the eliminations are independent, the actual features that were being used in those step
    could have been completely different. Using the selected number of features n, RFECV then performs another RFE, eliminating features
    until n features remain and returns these features as the result.
    """
    def __init__(self, cross_validator_params: VectorModelCrossValidatorParams, min_features=1):
        """
        :param cross_validator_params: the parameters for cross-validation
        :param min_features: the smallest number of features that shall be evaluated during feature elimination
        """
        if not cross_validator_params.returnTrainedModels:
            raise ValueError("crossValidatorParams: returnTrainedModels is required to be enabled")
        self.cross_validator_params = cross_validator_params
        self.min_features = min_features

    def run(self, model: Union[VectorModel, FeatureImportanceProvider], io_data: InputOutputData, metric_name: str,
            minimise: bool, remove_input_preprocessors=False) -> RFEResult:
        """
        Runs the optimisation for the given model and data.

        :param model: the model
        :param io_data: the data
        :param metric_name: the metric to optimise
        :param minimise: whether the metric shall be minimsed; if False, maximise.
        :param remove_input_preprocessors: whether to remove input preprocessors from the model and create input data
            only once during the entire experiment; this is usually reasonable only if all input preprocessors are not
            trained on the input data or if, for any given data split/fold, the preprocessor learning outcome is likely
            to be largely similar.
        :return: a result object, which provides access to the selected features and data on all elimination steps
        """
        metric_key = f"mean[{metric_name}]"

        if remove_input_preprocessors:
            model = copy(model)
            model.fit_input_output_data(io_data, fit_preprocessors=True, fit_model=False)
            inputs = model.compute_model_inputs(io_data.inputs)
            model.remove_input_preprocessors()
            io_data = InputOutputData(inputs, io_data.outputs)
            features = list(inputs.columns)
        else:
            features = None  # can only be obtained after having fitted the model initially (see below)
        dft_column_filter = DFTColumnFilter()
        model.with_feature_transformers(dft_column_filter, add=True)

        steps = []
        while True:
            # evaluate model
            cross_validator = create_vector_model_cross_validator(io_data, model=model, params=self.cross_validator_params)
            cross_val_data = cross_validator.eval_model(model)
            agg_metrics_dict = cross_val_data.get_eval_stats_collection().agg_metrics_dict()
            metric_value = agg_metrics_dict[metric_key]

            if features is None:
                features = cross_val_data.trained_models[0].get_model_input_variable_names()

            steps.append(RFEStep(metric_value=metric_value, features=features))

            # eliminate feature(s)
            log.info(f"Model performance with {len(features)} features: {metric_key}={metric_value}")
            agg_importance = AggregatedFeatureImportance(*cross_val_data.trained_models)
            fi = agg_importance.get_aggregated_feature_importance()
            tuples = fi.get_sorted_tuples()
            min_importance = tuples[0][1]
            if min_importance == 0:
                eliminated_features = []
                for i, (fname, importance) in enumerate(tuples):
                    if importance > 0:
                        break
                    eliminated_features.append(fname)
                log.info(f"Eliminating {len(eliminated_features)} features with 0 importance: {eliminated_features}")
            else:
                eliminated_features = [tuples[0][0]]
                log.info(f"Eliminating feature {eliminated_features[0]}")
            features = [f for f in features if f not in eliminated_features]
            dft_column_filter.keep = features

            log.info(f"{len(features)} features remain")

            if len(features) < self.min_features:
                log.info("Minimum number of features reached/exceeded")
                break

        return RFEResult(steps, metric_name, minimise)


class RecursiveFeatureElimination:
    def __init__(self, metric_computation: MetricComputation, min_features=1):
        """
        :param metric_computation: the method to apply for metric computation in order to determine which feature set is best
        :param min_features: the smallest number of features that shall be evaluated during feature elimination
        """
        self.metric_computation = metric_computation
        self.min_features = min_features

    def run(self, model_factory: Callable[[], Union[VectorRegressionModel, VectorClassificationModel]], minimise: bool) -> RFEResult:
        """
        Runs the optimisation for the given model and data.

        :param model_factory: factory for the model to be evaluated
        :param minimise: whether the metric shall be minimised; if False, maximise.
        :return: a result object, which provides access to the selected features and data on all elimination steps
        """
        features = None  # can only be obtained after having fitted the model initially (see below)
        dft_column_filter = DFTColumnFilter()  # kept features will be adapted in the loop below; added to each evaluated model

        steps = []
        while True:
            def create_model():
                return model_factory().with_feature_transformers(dft_column_filter, add=True)

            # compute metric
            metric_computation_result = self.metric_computation.compute_metric_value(create_model)
            metric_value = metric_computation_result.metric_value

            if features is None:
                # noinspection PyTypeChecker
                model: VectorModel = metric_computation_result.models[0]
                features = model.get_model_input_variable_names()

            steps.append(RFEStep(metric_value=metric_value, features=features))

            # eliminate feature(s)
            log.info(f"Model performance with {len(features)} features: metric={metric_value}")
            agg_importance = AggregatedFeatureImportance(*metric_computation_result.models)
            fi = agg_importance.get_aggregated_feature_importance()
            tuples = fi.get_sorted_tuples()
            min_importance = tuples[0][1]
            if min_importance == 0:
                eliminated_features = []
                for i, (fname, importance) in enumerate(tuples):
                    if importance > 0:
                        break
                    eliminated_features.append(fname)
                log.info(f"Eliminating {len(eliminated_features)} features with 0 importance: {eliminated_features}")
            else:
                eliminated_features = [tuples[0][0]]
                log.info(f"Eliminating feature {eliminated_features[0]}")
            features = [f for f in features if f not in eliminated_features]
            dft_column_filter.keep = features

            log.info(f"{len(features)} features remain")

            if len(features) < self.min_features:
                log.info("Minimum number of features reached/exceeded")
                break

        return RFEResult(steps, self.metric_computation.metric.name, minimise)



import collections
import datetime
import logging
import typing
from abc import ABC, abstractmethod
from typing import Callable, List, Iterable, Optional

import numpy as np
import pandas as pd

from . import util, data_transformation
from .distance_metric import DistanceMetric
from .featuregen import FeatureGeneratorFromNamedTuples
from .util.string import object_repr
from .util.typing import PandasNamedTuple
from .vector_model import VectorClassificationModel, VectorRegressionModel

log = logging.getLogger(__name__)


class Neighbor:
    def __init__(self, value: PandasNamedTuple, distance: float):
        self.distance = distance
        self.value = value
        self.identifier = value.Index


class NeighborProvider(ABC):
    def __init__(self, df_indexed_by_id: pd.DataFrame):
        self.df = df_indexed_by_id
        self.index = self.df.index
        if any(self.index.duplicated()):
            raise Exception("Dataframe index should not contain duplicates")
        self.index_position_dict = {idx: pos for pos, idx in enumerate(self.index)}

    @abstractmethod
    def iter_potential_neighbors(self, value: PandasNamedTuple) -> Iterable[PandasNamedTuple]:
        pass

    @abstractmethod
    def __str__(self):
        return super().__str__()


class AllNeighborsProvider(NeighborProvider):
    def __init__(self, df_indexed_by_id: pd.DataFrame):
        super().__init__(df_indexed_by_id)
        self.named_tuples = None

    def __getstate__(self):
        d = self.__dict__.copy()
        d["namedTuples"] = None
        return d

    def iter_potential_neighbors(self, value):
        identifier = value.Index
        if self.named_tuples is None:
            self.named_tuples = list(self.df.itertuples())
        for nt in self.named_tuples:
            if nt.Index != identifier:
                yield nt

    def __str__(self):
        return str(self.__class__.__name__)


class TimerangeNeighborsProvider(NeighborProvider):
    def __init__(self, df_indexed_by_id: pd.DataFrame, timestamps_column="timestamps",
                 past_time_range_days=120, future_time_range_days=120):
        super().__init__(df_indexed_by_id)
        if not pd.core.dtypes.common.is_datetime64_any_dtype(self.df[timestamps_column]):
            raise Exception(f"Column {timestamps_column} does not have a compatible datatype")
        self.timestamps_column = timestamps_column
        self.past_time_range_days = past_time_range_days
        self.future_time_range_days = future_time_range_days
        self.past_timedelta = datetime.timedelta(days=past_time_range_days)
        self.future_timedelta = datetime.timedelta(days=future_time_range_days)

    def iter_potential_neighbors(self, value: PandasNamedTuple):
        identifier = value.Index
        input_time = getattr(value, self.timestamps_column)
        max_time, min_time = input_time + self.future_timedelta, input_time - self.past_timedelta
        neighbors_df = self.df[
            self.df[self.timestamps_column].apply(lambda time: min_time < time < input_time)
        ]
        if identifier in neighbors_df.index:
            neighbors_df.drop(identifier, inplace=True)
        return neighbors_df.itertuples()

    def __str__(self):
        return object_repr(self, ["past_time_range_days", "future_time_range_days"])


class AbstractKnnFinder(ABC):
    @abstractmethod
    def find_neighbors(self, named_tuple: PandasNamedTuple, n_neighbors=20) -> List[Neighbor]:
        pass

    @abstractmethod
    def __str__(self):
        super().__str__()


class CachingKNearestNeighboursFinder(AbstractKnnFinder):
    """
    A nearest neighbor finder which uses a cache for distance metrics in order speed up repeated computations
    of the neighbors of the same data point by keeping a pandas.Series of distances to all provided
    data points cached. If the distance metric is of the composite type LinearCombinationDistanceMetric,
    its component distance metrics are cached, such that weights in the linear combination can be varied
    without necessitating recomputations.
    """
    log = log.getChild(__qualname__)

    def __init__(self, cache: 'CachingKNearestNeighboursFinder.DistanceMetricCache', distance_metric: DistanceMetric,
            neighbor_provider: NeighborProvider):
        self.neighbor_provider = neighbor_provider
        # This field is purely for logging purposes
        self.distance_metric = distance_metric
        if isinstance(distance_metric, distance_metric.LinearCombinationDistanceMetric):
            self.weighted_distance_metrics = [(cache.get_cached_metric(dm), w) for (w, dm) in distance_metric.metrics]
        else:
            self.weighted_distance_metrics = [(cache.get_cached_metric(distance_metric), 1)]

    def __str__(self):
        return object_repr(self, ["neighbor_provider", "distance_metric"])

    class DistanceMetricCache:
        """
        A cache for distance metrics which identifies equivalent distance metrics by their string representations.
        The cache can be passed (consecutively) to multiple KNN models in order to speed up computations for the
        same test data points. If the cache is reused, it is assumed that the neighbor provider remains the same.
        """
        log = log.getChild(__qualname__)

        def __init__(self):
            self._cached_metrics_by_name = {}

        def get_cached_metric(self, distance_metric):
            key = str(distance_metric)
            cached_metric = self._cached_metrics_by_name.get(key)
            if cached_metric is None:
                self.log.info(f"Creating new cached metric for key '{key}'")
                cached_metric = CachingKNearestNeighboursFinder.CachedSeriesDistanceMetric(distance_metric)
                self._cached_metrics_by_name[key] = cached_metric
            else:
                self.log.info(f"Reusing cached metric for key '{key}'")
            return cached_metric

    class CachedSeriesDistanceMetric:
        """
        Provides caching for a wrapped distance metric: the series of all distances to provided potential neighbors
        are retained in a cache
        """
        def __init__(self, distance_metric):
            self.distance_metric = distance_metric
            self.cache = {}

        def get_distance_series(self, named_tuple: PandasNamedTuple, potential_neighbor_values):
            identifier = named_tuple.Index
            series = self.cache.get(identifier)
            if series is None:
                distances = []
                for neighborTuple in potential_neighbor_values:
                    distances.append(self.distance_metric.distance(named_tuple, neighborTuple))
                series = pd.Series(distances)
                self.cache[identifier] = series
            return series

    def find_neighbors(self, named_tuple: PandasNamedTuple, n_neighbors=20) -> List[Neighbor]:
        potential_neighbors = list(self.neighbor_provider.iter_potential_neighbors(named_tuple))
        summed_distance_series = None
        for i, (metric, weight) in enumerate(self.weighted_distance_metrics):
            weighted_distances_series = metric.get_distance_series(named_tuple, potential_neighbors) * weight
            if i == 0:
                summed_distance_series = weighted_distances_series.copy()
            else:
                summed_distance_series += weighted_distances_series
        summed_distance_series.sort_values(ascending=True, inplace=True)
        result = []
        for i in range(n_neighbors):
            neighbor_tuple = potential_neighbors[summed_distance_series.index[i]]
            distance = summed_distance_series.iloc[i]
            result.append(Neighbor(neighbor_tuple, distance))
        return result


class KNearestNeighboursFinder(AbstractKnnFinder):
    def __init__(self, distance_metric: DistanceMetric, neighbor_provider: NeighborProvider):
        self.neighbor_provider = neighbor_provider
        self.distance_metric = distance_metric

    def __str__(self):
        return object_repr(self, ["neighbor_provider", "distance_metric"])

    def find_neighbors(self, named_tuple: PandasNamedTuple, n_neighbors=20) -> List[Neighbor]:
        result = []
        log.debug(f"Finding neighbors for {named_tuple.Index}")
        for neighborTuple in self.neighbor_provider.iter_potential_neighbors(named_tuple):
            distance = self.distance_metric.distance(named_tuple, neighborTuple)
            result.append(Neighbor(neighborTuple, distance))
        result.sort(key=lambda n: n.distance)
        return result[:n_neighbors]


class KNearestNeighboursClassificationModel(VectorClassificationModel):
    def __init__(self, num_neighbors: int, distance_metric: DistanceMetric,
            neighbor_provider_factory: Callable[[pd.DataFrame], NeighborProvider] = AllNeighborsProvider,
            distance_based_weighting=False, distance_epsilon=1e-3,
            distance_metric_cache: CachingKNearestNeighboursFinder.DistanceMetricCache = None, **kwargs):
        """
        :param num_neighbors: the number of nearest neighbors to consider
        :param distance_metric: the distance metric to use
        :param neighbor_provider_factory: a factory with which a neighbor provider can be constructed using data
        :param distance_based_weighting: whether to weight neighbors according to their distance (inverse); if False, use democratic vote
        :param distance_epsilon: a distance that is added to all distances for distance-based weighting (in order to avoid 0 distances);
        :param distance_metric_cache: a cache for distance metrics which shall be used to store speed up repeated computations
            of the neighbors of the same data point by keeping series of distances cached (particularly for composite distance metrics);
            see class CachingKNearestNeighboursFinder
        :param kwargs: parameters to pass on to super-classes
        """
        super().__init__(**kwargs)
        self.distance_epsilon = distance_epsilon
        self.distance_based_weighting = distance_based_weighting
        self.neighbor_provider_factory = neighbor_provider_factory
        self.num_neighbors = num_neighbors
        self.distance_metric = distance_metric
        self.distance_metric_cache = distance_metric_cache
        self.df = None
        self.y = None
        self.knn_finder = None

    def _tostring_excludes(self) -> List[str]:
        return super()._tostring_excludes() + ["neighbor_provider_factory", "distance_metric", "distance_metric_cache", "df", "y"]

    # noinspection DuplicatedCode
    def _fit_classifier(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None):
        self._warn_sample_weights_unsupported(False, weights)
        assert len(y.columns) == 1, "Expected exactly one column in label set Y"
        self.df = x.merge(y, how="inner", left_index=True, right_index=True)
        self.y = y
        neighbor_provider = self.neighbor_provider_factory(self.df)
        if self.distance_metric_cache is None:
            self.knn_finder = KNearestNeighboursFinder(self.distance_metric, neighbor_provider)
        else:
            self.knn_finder = CachingKNearestNeighboursFinder(self.distance_metric_cache, self.distance_metric, neighbor_provider)
        log.info(f"Using neighbor provider of type {self.knn_finder.__class__.__name__}")

    def _predict_class_probabilities(self, x: pd.DataFrame):
        output_df = pd.DataFrame({label: np.nan for label in self._labels}, index=x.index)
        for nt in x.itertuples():
            neighbors = self.find_neighbors(nt)
            probabilities = self._predict_class_probability_vector_from_neighbors(neighbors)
            output_df.loc[nt.Index] = probabilities
        return output_df

    def _predict_class_probability_vector_from_neighbors(self, neighbors: List['Neighbor']):
        weights = collections.defaultdict(lambda: 0)
        total = 0
        for neigh in neighbors:
            if self.distance_based_weighting:
                weight = 1.0 / (neigh.distance + self.distance_epsilon)
            else:
                weight = 1
            weights[self._get_label(neigh)] += weight
            total += weight
        return [weights[label] / total for label in self._labels]

    def _get_label(self, neighbor: 'Neighbor'):
        return self.y.iloc[:, 0].loc[neighbor.identifier]

    def find_neighbors(self, named_tuple):
        return self.knn_finder.find_neighbors(named_tuple, self.num_neighbors)


class KNearestNeighboursRegressionModel(VectorRegressionModel):
    def __init__(self, num_neighbors: int, distance_metric: DistanceMetric,
            neighbor_provider_factory: Callable[[pd.DataFrame], NeighborProvider] = AllNeighborsProvider,
            distance_based_weighting=False, distance_epsilon=1e-3,
            distance_metric_cache: CachingKNearestNeighboursFinder.DistanceMetricCache = None, **kwargs):
        """
        :param num_neighbors: the number of nearest neighbors to consider
        :param distance_metric: the distance metric to use
        :param neighbor_provider_factory: a factory with which a neighbor provider can be constructed using data
        :param distance_based_weighting: whether to weight neighbors according to their distance (inverse); if False, use democratic vote
        :param distance_epsilon: a distance that is added to all distances for distance-based weighting (in order to avoid 0 distances);
        :param distance_metric_cache: a cache for distance metrics which shall be used to store speed up repeated computations
            of the neighbors of the same data point by keeping series of distances cached (particularly for composite distance metrics);
            see class CachingKNearestNeighboursFinder
        :param kwargs: parameters to pass on to super-classes
        """
        super().__init__(**kwargs)
        self.distance_epsilon = distance_epsilon
        self.distance_based_weighting = distance_based_weighting
        self.neighbor_provider_factory = neighbor_provider_factory
        self.num_neighbors = num_neighbors
        self.distance_metric = distance_metric
        self.distance_metric_cache = distance_metric_cache
        self.df = None
        self.y = None
        self.knn_finder = None

    def _tostring_excludes(self) -> List[str]:
        return super()._tostring_excludes() + ["neighbor_provider_factory", "distance_metric", "distance_metric_cache", "df", "y"]

    # noinspection DuplicatedCode
    def _fit(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None):
        self._warn_sample_weights_unsupported(False, weights)
        assert len(y.columns) == 1, "Expected exactly one column in label set Y"
        self.df = x.merge(y, how="inner", left_index=True, right_index=True)
        self.y = y
        neighbor_provider = self.neighbor_provider_factory(self.df)
        if self.distance_metric_cache is None:
            self.knn_finder = KNearestNeighboursFinder(self.distance_metric, neighbor_provider)
        else:
            self.knn_finder = CachingKNearestNeighboursFinder(self.distance_metric_cache, self.distance_metric, neighbor_provider)
        log.info(f"Using neighbor provider of type {self.knn_finder.__class__.__name__}")

    def _get_target(self, neighbor: Neighbor):
        return self.y.iloc[:, 0].loc[neighbor.identifier]

    def _predict_single_input(self, named_tuple):
        neighbors = self.knn_finder.find_neighbors(named_tuple, self.num_neighbors)
        neighbor_targets = np.array([self._get_target(n) for n in neighbors])
        if self.distance_based_weighting:
            neighbor_weights = np.array([1.0 / (n.distance + self.distance_epsilon) for n in neighbors])
            return np.sum(neighbor_targets * neighbor_weights) / np.sum(neighbor_weights)
        else:
            return np.mean(neighbor_targets)

    def _predict(self, x: pd.DataFrame) -> pd.DataFrame:
        predicted_values = []
        for i, nt in enumerate(x.itertuples()):
            predicted_values.append(self._predict_single_input(nt))
        return pd.DataFrame({self._predictedVariableNames[0]: predicted_values}, index=x.index)


class FeatureGeneratorNeighbors(FeatureGeneratorFromNamedTuples):
    """
    Generates features based on nearest neighbors. For each neighbor, a set of features is added to the output data frame.
    Each feature has the name "n{0-based neighbor index}_{feature name}", where the feature names are configurable
    at construction. The feature name "distance", which indicates the distance of the neighbor to the data point is
    always present.
    """
    def __init__(self, num_neighbors: int,
            neighbor_attributes: typing.List[str],
            distance_metric: DistanceMetric,
            neighbor_provider_factory: typing.Callable[[pd.DataFrame], NeighborProvider] = AllNeighborsProvider,
            cache: util.cache.KeyValueCache = None,
            categorical_feature_names: typing.Sequence[str] = (),
            normalisation_rules: typing.Sequence[data_transformation.DFTNormalisation.Rule] = ()):
        """
        :param num_neighbors: the number of neighbors for to generate features
        :param neighbor_attributes: the attributes of the neighbor's named tuple to include as features (in addition to "distance")
        :param distance_metric: the distance metric defining which neighbors are near
        :param neighbor_provider_factory: a factory for the creation of neighbor provider
        :param cache: an optional key-value cache in which feature values are stored by data point identifier (as given by the DataFrame's
            index)
        """
        super().__init__(cache=cache, categorical_feature_names=categorical_feature_names, normalisation_rules=normalisation_rules)
        self.neighbor_attributes = neighbor_attributes
        self.distance_metric = distance_metric
        self.neighbor_provider_factory = neighbor_provider_factory
        self.num_neighbors = num_neighbors
        self._knn_finder: Optional[KNearestNeighboursFinder] = None
        self._train_x = None

    def _generate(self, df: pd.DataFrame, ctx=None):
        if self._train_x is None:
            raise Exception("Feature generator has not been fitted")
        neighbor_provider = self.neighbor_provider_factory(self._train_x)
        self._knn_finder = KNearestNeighboursFinder(self.distance_metric, neighbor_provider)
        return super()._generate(df, ctx)

    def _generate_feature_dict(self, named_tuple) -> typing.Dict[str, typing.Any]:
        neighbors = self._knn_finder.find_neighbors(named_tuple, self.num_neighbors)
        result = {}
        for i, neighbor in enumerate(neighbors):
            result[f"n{i}_distance"] = neighbor.distance
            for attr in self.neighbor_attributes:
                result[f"n{i}_{attr}"] = getattr(neighbor.value, attr)
        return result

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame = None, ctx=None):
        self._train_x = x



from __future__ import annotations
import collections
import functools
import pickle
import sys
from abc import ABC, abstractmethod
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Callable, Dict, Union, Any, List, Sequence, Optional
import json
import logging
import re
import threading


from azure.storage.table import TableService, TableBatch, Entity
from azure.storage.blob import BlockBlobService
import pandas as pd
import numpy as np

from .cache import PersistentKeyValueCache, PeriodicUpdateHook

_log = logging.getLogger(__name__)


class Serialiser(ABC):
    """
    Abstraction for mechanisms to serialise values, which do not fit table storage data model,
    see https://docs.microsoft.com/en-us/rest/api/storageservices/understanding-the-table-service-data-model
    """

    @abstractmethod
    def serialise(self, value) -> str:
        pass

    @abstractmethod
    def deserialise(self, value: str):
        pass


class NumpyArrayJsonSerialiser(Serialiser):
    """
    Serialises a numpy array as json string of list representation of array
    """

    def serialise(self, value: np.ndarray) -> str:
        return json.dumps(value.tolist())

    def deserialise(self, value: str):
        return np.array(json.loads(value))


class PropertyLoader(ABC):
    """
    Abstraction of a customised loader for an entity property
    """

    @abstractmethod
    def load_property_value(self, entity: Entity):
        pass

    @abstractmethod
    def write_property_value(self, entity: Entity):
        pass

    @abstractmethod
    def load_property_value_to_data_frame_column(self, df: pd.DataFrame):
        pass


class SerialisedPropertyLoader(PropertyLoader):
    """
    PropertyLoader to serialise and de-serialise values. Useful, if type of values is not aligned with table storage data model,
    see https://docs.microsoft.com/en-us/rest/api/storageservices/understanding-the-table-service-data-model
    """
    def __init__(self, property_name: str, serialiser: Serialiser):
        self.serialiser = serialiser
        self.property_name = property_name

    def load_property_value(self, entity: Entity):
        entity[self.property_name] = self.serialiser.deserialise(entity[self.property_name])

    def write_property_value(self, entity: Entity):
        entity[self.property_name] = self.serialiser.serialise(entity[self.property_name])

    def load_property_value_to_data_frame_column(self, df: pd.DataFrame):
        if self.property_name in df.columns:
            df.loc[:, self.property_name] = [self.serialiser.deserialise(value) for value in df[self.property_name]]


class AzureTableBlobBackend(ABC):
    """
    Abstraction of a blob backend, which allows for convenient setting and getting of values stored in blob storage via a
    reference to the value
    """

    @abstractmethod
    def get_value_from_reference(self, value_identifier: str):
        pass

    @abstractmethod
    def get_value_reference(self, partition_key: str, row_key: str, value_name: str, blob_name_prefix: str = None) -> str:
        pass

    @abstractmethod
    def set_value_for_reference(self, value_identifier: str, value):
        pass


class BlobPerKeyAzureTableBlobBackend(AzureTableBlobBackend, ABC):

    """
    Backend stores serialised values as /tableName/partitionKey/rowKey/valueName.<fileExtension>
    or /tableName/rowKey/valueName.<fileExtension>, if partitionKey equals tableName
    """

    def __init__(self, block_blob_service: BlockBlobService, container_name: str):
        """

        :param block_blob_service: https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.blockblobservice.blockblobservice?view=azure-python-previous
        """
        self.block_blob_service = block_blob_service
        self.container_name = container_name
        self.container_list = [container.name for container in block_blob_service.list_containers()]
        if container_name not in self.container_list:
            self.block_blob_service.create_container(container_name)
            self.container_list.append(container_name)

    @property
    @abstractmethod
    def file_extension(self):
        pass

    @abstractmethod
    def _get_blob_value(self, container_name, blob_name):
        pass

    @abstractmethod
    def _write_value_to_blob(self, container_name, blob_name, value):
        pass

    def get_value_from_reference(self, value_identifier: str):
        container_name = self._get_container_name_from_identifier(value_identifier)
        blob_name = self._get_blob_name_from_identifier(value_identifier)
        return self._get_blob_value(container_name, blob_name)

    def get_value_reference(self, partition_key: str, row_key: str, value_name: str, blob_name_prefix: str = None) -> str:
        blob_name = self._get_blob_name_from_keys(partition_key, row_key, value_name, blob_prefix=blob_name_prefix)
        return self.block_blob_service.make_blob_url(self.container_name, blob_name)

    def set_value_for_reference(self, value_identifier: str, value):
        container_name = self._get_container_name_from_identifier(value_identifier)
        blob_name = self._get_blob_name_from_identifier(value_identifier)
        self._write_value_to_blob(container_name, blob_name, value)

    def _get_blob_name_from_identifier(self, value_identifier: str):
        return (value_identifier.partition(f"{self.block_blob_service.primary_endpoint}/")[2]).partition("/")[2]

    def _get_container_name_from_identifier(self, value_identifier: str):
        return (value_identifier.partition(f"{self.block_blob_service.primary_endpoint}/")[2]).partition("/")[0]

    def _get_blob_name_from_keys(self, partition_key: str, row_key: str, value_name: str, blob_prefix: str = None):
        identifier_list = [blob_prefix, partition_key] if blob_prefix is not None and blob_prefix != partition_key else [partition_key]
        identifier_list.extend([row_key, value_name])
        return "/".join(identifier_list) + self.file_extension


class TextDumpAzureTableBlobBackend(BlobPerKeyAzureTableBlobBackend):
    """
   Backend stores values as txt files in the structure /tableName/partitionKey/rowKey/valueName
   """

    @property
    def file_extension(self):
        return ""

    def _get_blob_value(self, container_name, blob_name):
        return self.block_blob_service.get_blob_to_text(container_name, blob_name).content

    def _write_value_to_blob(self, container_name, blob_name, value):
        self.block_blob_service.create_blob_from_text(container_name, blob_name, value)


class JsonAzureTableBlobBackend(BlobPerKeyAzureTableBlobBackend):
    """
    Backend stores values as json files in the structure /tableName/partitionKey/rowKey/valueName.json
    """

    @property
    def file_extension(self):
        return ".json"

    def _get_blob_value(self, container_name, blob_name):
        encoded_value = self.block_blob_service.get_blob_to_bytes(container_name, blob_name).content
        return self._decode_bytes_to_value(encoded_value)

    def _write_value_to_blob(self, container_name, blob_name, value):
        encoded_value = self._encode_value_to_bytes(value)
        self.block_blob_service.create_blob_from_bytes(container_name, blob_name, encoded_value)

    @staticmethod
    def _encode_value_to_bytes(value):
        return str.encode(json.dumps(value))

    @staticmethod
    def _decode_bytes_to_value(_bytes):
        return json.loads(_bytes.decode())


class PickleAzureTableBlobBackend(JsonAzureTableBlobBackend):
    """
    Backend stores values as pickle files in the structure /tableName/partitionKey/rowKey/valueName.pickle
    """

    @property
    def file_extension(self):
        return ".pickle"

    @staticmethod
    def _encode_value_to_bytes(value):
        return pickle.dumps(value)

    @staticmethod
    def _decode_bytes_to_value(_bytes):
        return pickle.loads(_bytes)


class BlobBackedPropertyLoader(PropertyLoader):
    AZURE_ALLOWED_SIZE_PER_PROPERTY_BYTES = 64000
    AZURE_ALLOWED_STRING_LENGTH_PER_PROPERTY = 32000

    """
    PropertyLoader to write and read values from blob backend via a reference to the value. Useful, if values cannot
    be stored in table storage itself, due to not being aligned with table storage data model, 
    see https://docs.microsoft.com/en-us/rest/api/storageservices/understanding-the-table-service-data-model
    """
    def __init__(self, property_name: str, blob_backend: AzureTableBlobBackend, blob_prefix: str = None,
            property_boolean_blob_status_name: str = None, max_workers=None):
        """
        :param property_name: name of property in table
        :param property_boolean_blob_status_name: name of property representing a boolean flag within a table, which indicates, if value is
            blob backed. If None, each value is assumed to be blob backed.
        :param blob_backend: actual backend to use for storage
        :param blob_prefix: prefix to use for blob in storage, e.g. a table name
        :param max_workers: maximal number of workers to load data from blob storage
        """
        self.blob_prefix = blob_prefix
        self.property_blob_status_name = property_boolean_blob_status_name
        self.blob_backend = blob_backend
        self.max_workers = max_workers
        self.propertyName = property_name

    def load_property_value(self, entity: Entity):
        if self._is_entity_value_blob_backed(entity):
            entity[self.propertyName] = self.blob_backend.get_value_from_reference(entity[self.propertyName])

    def write_property_value(self, entity: Entity):
        if self.propertyName in entity.keys():
            if self._need_to_write_to_blob(entity[self.propertyName]):
                value_identifier = self.blob_backend.get_value_reference(entity["PartitionKey"], entity["RowKey"], self.propertyName,
                    blob_name_prefix=self.blob_prefix)
                value = entity[self.propertyName]
                self.blob_backend.set_value_for_reference(value_identifier, value)
                entity[self.propertyName] = value_identifier
                property_blob_status = True if self.property_blob_status_name is not None else None
            else:
                property_blob_status = False if self.property_blob_status_name is not None else None

            if property_blob_status is not None:
                entity[self.property_blob_status_name] = property_blob_status

    def load_property_value_to_data_frame_column(self, df: pd.DataFrame):
        if self.propertyName in df.columns:
            if self.property_blob_status_name is None:
                df.loc[:, self.propertyName] = self._load_values_in_series(df[self.propertyName])
            else:
                df.loc[df[self.property_blob_status_name], self.propertyName] = \
                    self._load_values_in_series(df.loc[df[self.property_blob_status_name], self.propertyName])

    def _need_to_write_to_blob(self, value):
        if self.property_blob_status_name is None:
            return True
        if sys.getsizeof(value) > self.AZURE_ALLOWED_SIZE_PER_PROPERTY_BYTES:
            return True
        if isinstance(value, str) and len(value) > self.AZURE_ALLOWED_STRING_LENGTH_PER_PROPERTY:
            return True
        return False

    def _is_entity_value_blob_backed(self, entity: Entity):
        if self.propertyName not in entity.keys():
            return False
        if self.property_blob_status_name is None or self.property_blob_status_name not in entity:
            return True
        return entity[self.property_blob_status_name]

    def _load_values_in_series(self, _series: pd.Series):
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            _series = list(executor.map(self.blob_backend.get_value_from_reference, _series))
        return _series


class BlobBackedSerialisedPropertyLoader(BlobBackedPropertyLoader, SerialisedPropertyLoader):
    """
    Property loader, which combines serialisation and blob backing.
    """
    def __init__(self, property_name, serialiser: Serialiser, blob_backend: AzureTableBlobBackend, blob_prefix: str = None,
            property_boolean_blob_status_name: str = None, max_workers=None):
        """


        :param property_name: name of property in table
        :param serialiser:
        :param property_boolean_blob_status_name: name of property representing a boolean flag within a table, which indicates, if value is
            blob backed. If None, each value is assumed to be blob backed.
        :param blob_backend: actual backend to use for storage
        :param blob_prefix: prefix to use for blob in storage, e.g. a table name
        :param max_workers: maximal number of workers to load data from blob storage
        """
        SerialisedPropertyLoader.__init__(self, property_name, serialiser)
        BlobBackedPropertyLoader.__init__(self, property_name, blob_backend, blob_prefix, property_boolean_blob_status_name, max_workers)

    def load_property_value(self, entity: Entity):
        super(BlobBackedPropertyLoader, self).load_property_value(entity)
        super(SerialisedPropertyLoader, self).load_property_value(entity)

    def write_property_value(self, entity: Entity):
        super(SerialisedPropertyLoader, self).write_property_value(entity)
        super(BlobBackedPropertyLoader, self).write_property_value(entity)

    def load_property_value_to_data_frame_column(self, df: pd.DataFrame):
        super(BlobBackedPropertyLoader, self).load_property_value_to_data_frame_column(df)
        super(SerialisedPropertyLoader, self).load_property_value_to_data_frame_column(df)


class AzureLazyBatchCommitTable:
    """
    Wrapper for an Azure table, which allow for convenient insertion via lazy batch execution per partition.
    Uses a priority queue to manage order of partitions to be committed.
    To execute insertions, call :func:`LazyBatchCommitTable.commit`
    """

    AZURE_ALLOWED_TABLE_NAME_PATTERN = re.compile("^[A-Za-z][A-Za-z0-9]{2,62}$")
    AZURE_ALLOWED_TABLE_BATCH_SIZE = 100

    class PartitionCommandsPriorityQueue:

        class PartitionCommands:
            def __init__(self, partition_key):
                self.partition_key = partition_key
                self._command_list = collections.deque()

            def __len__(self):
                return len(self._command_list)

            def append(self, command):
                self._command_list.append(command)

            def execute(self, context_manager: Callable[[], TableBatch], batch_size: int):
                while len(self._command_list) > 0:
                    _slice = [self._command_list.popleft() for _ in range(min(batch_size, len(self._command_list)))]
                    _log.info(f"Committing {len(_slice)} cache entries to the database")
                    with context_manager() as batch:
                        for command in _slice:
                            command(batch)

        def __init__(self):
            self.partition_commands_queue = []
            self.partition_key2_commands = {}
            self._thread_lock = threading.Lock()

        def add_command(self, partition_key, command: Union[Callable[[TableBatch], Any], functools.partial[TableBatch]]):
            """
            Add a command to queue of corresponding partitionKey
            :param partition_key:
            :param command: a callable on a TableBatch
            """
            with self._thread_lock:
                if partition_key not in self.partition_key2_commands:
                    commands = self.PartitionCommands(partition_key)
                    self.partition_commands_queue.append(commands)
                    self.partition_key2_commands[partition_key] = commands
                self.partition_key2_commands[partition_key].append(command)

        def pop(self, min_length: int = None) -> Optional[AzureLazyBatchCommitTable.PartitionCommandsPriorityQueue.PartitionCommands]:
            """
            :param min_length: minimal length of largest PartitionCommands for the pop to take place.
            :return: largest PartitionCommands or None if minimal length is not reached
            """
            with self._thread_lock:
                return self._pop(min_length)

        def pop_all(self):
            with self._thread_lock:
                commands_list = []
                while not self._is_empty():
                    commands_list.append(self._pop())
                return commands_list

        def is_empty(self):
            with self._thread_lock:
                return self._is_empty()

        def _pop(self, min_length=None):
            length, index = self._get_max_priority_info()
            if index is not None and (min_length is None or length >= min_length):
                q = self.partition_commands_queue.pop(index)
                del self.partition_key2_commands[q.partition_key]
                return q
            else:
                return None

        def _is_empty(self):
            return len(self.partition_commands_queue) == 0

        def _get_max_priority_info(self):
            lengths_list = list(map(len, self.partition_commands_queue))
            if len(lengths_list) == 0:
                return 0, None
            max_length = max(lengths_list)
            return max_length, lengths_list.index(max_length)

    def __init__(self, table_name: str, table_service: TableService, property_loaders: Sequence[PropertyLoader] = ()):
        """
        :param table_name: name of table
        :param table_service: instance of :class:`azure.storage.table.TableService` to connect to Azure table storage
        :param property_loaders:
        """

        if not self.AZURE_ALLOWED_TABLE_NAME_PATTERN.match(table_name):
            raise ValueError(f"Invalid table name {table_name}, see: "
                             f"https://docs.microsoft.com/en-us/rest/api/storageservices/Understanding-the-Table-Service-Data-Model")

        self.table_service = table_service
        self.table_name = table_name
        self.property_loaders = property_loaders
        self._partition_queues = self.PartitionCommandsPriorityQueue()
        self._context_manager = functools.partial(self.table_service.batch, self.table_name)

        if not self.exists():
            self.table_service.create_table(self.table_name)

    def insert_or_replace_entity(self, entity: Union[Dict, Entity]):
        """
        Lazy wrapper method for :func:`azure.storage.table.TableService.insert_or_replace_entity`
        :param entity:
        """
        partition_key = entity["PartitionKey"]
        for property_loader in self.property_loaders:
            property_loader.write_property_value(entity)
        execution_command = functools.partial(self._insert_or_replace_entity_via_batch, entity)
        self._partition_queues.add_command(partition_key, execution_command)

    def insert_entity(self, entity: Union[Dict, Entity]):
        """
        Lazy wrapper method for :func:`azure.storage.table.TableService.insert_entity`
        :param entity:
        """
        partition_key = entity["PartitionKey"]
        for property_loader in self.property_loaders:
            property_loader.write_property_value(entity)
        execution_command = functools.partial(self._insert_entity_via_batch, entity)
        self._partition_queues.add_command(partition_key, execution_command)

    def get_entity(self, partition_key: str, row_key: str) -> Optional[Entity]:
        """
        Wraps :func:`azure.storage.table.TableService.get_entity`
        :param partition_key:
        :param row_key:
        :return:
        """
        try:
            entity = self.table_service.get_entity(self.table_name, partition_key, row_key)
            for property_loader in self.property_loaders:
                property_loader.load_property_value(entity)
            return entity
        except Exception as e:
            _log.debug(f"Unable to load value for partitionKey {partition_key} and rowKey {row_key} from table {self.table_name}: {e}")
            return None

    def commit_blocking_until_empty(self, max_batch_size=AZURE_ALLOWED_TABLE_BATCH_SIZE):
        """
        Commit insertion commands. Commands are executed batch-wise per partition until partition queue is empty in a
        blocking manner.
        :param max_batch_size: maximal batch size to use for batch insertion, must be less or equal to batch size allowed by Azure
        """

        max_batch_size = self._validate_max_batch_size(max_batch_size)

        while not self._partition_queues.is_empty():
            commands = self._partition_queues.pop()
            commands.execute(self._context_manager, max_batch_size)

    def commit_non_blocking_current_queue_state(self, max_batch_size=AZURE_ALLOWED_TABLE_BATCH_SIZE):
        """
        Commit insertion commands. Empties the current PartitionCommandsQueue in a non blocking way.
        Commands are executed batch-wise per partition.
        :param max_batch_size: maximal batch size to use for batch insertion, must be less or equal to batch size allowed by Azure
        """

        max_batch_size = self._validate_max_batch_size(max_batch_size)

        def commit():
            commands_list = self._partition_queues.pop_all()
            for commands in commands_list:
                commands.execute(self._context_manager, max_batch_size)

        thread = threading.Thread(target=commit, daemon=False)
        thread.start()

    def commit_blocking_largest_partition_from_queue(self, max_batch_size=AZURE_ALLOWED_TABLE_BATCH_SIZE, min_length=None):
        """
        Commits in a blocking way the largest partition from PartitionCommandsQueue
        :param max_batch_size: maximal batch size to use for batch insertion, must be less or equal to batch size allowed by Azure
        :param min_length: minimal size of largest partition. If not None, pop and commit only if minLength is reached.
        :return:
        """
        max_batch_size = self._validate_max_batch_size(max_batch_size)
        commands = self._partition_queues.pop(min_length)
        if commands is not None:
            commands.execute(self._context_manager, max_batch_size)

    def _validate_max_batch_size(self, max_batch_size):
        if max_batch_size > self.AZURE_ALLOWED_TABLE_BATCH_SIZE:
            _log.warning(f"Provided maxBatchSize is larger than allowed size {self.AZURE_ALLOWED_TABLE_BATCH_SIZE}. "
                         f"Will use maxBatchSize {self.AZURE_ALLOWED_TABLE_BATCH_SIZE} instead.")
            max_batch_size = self.AZURE_ALLOWED_TABLE_BATCH_SIZE
        return max_batch_size

    def load_table_to_data_frame(self, columns: List[str] = None, row_filter_query: str = None, num_records: int = None):
        """
        Load all rows of table to :class:`~pandas.DataFrame`
        :param row_filter_query:
        :param num_records:
        :param columns: restrict loading to provided columns
        :return: :class:`~pandas.DataFrame`
        """
        if num_records is None:
            records = list(self._iter_records(columns, row_filter_query))
        else:
            records = []
            for record in self._iter_records(columns, row_filter_query):
                records.append(record)
                if len(records) >= num_records:
                    break
        df = pd.DataFrame(records, columns=columns)
        for property_loader in self.property_loaders:
            property_loader.load_property_value_to_data_frame_column(df)
        return df

    def iter_data_frame_chunks(self, chunk_size: int, columns: List[str] = None, row_filter_query: str = None):
        """
        Get a generator of dataframe chunks
        :param row_filter_query:
        :param chunk_size:
        :param columns:
        :return:
        """
        records = []
        for record in self._iter_records(columns, row_filter_query):
            records.append(record)
            if len(records) >= chunk_size:
                df = pd.DataFrame(records, columns=columns)
                for propertyLoader in self.property_loaders:
                    propertyLoader.load_property_value_to_data_frame_column(df)
                yield df
                records = []

    def iter_records(self, columns: List[str] = None, row_filter_query: str = None):
        """

        Get a generator of table entities
        :param row_filter_query:
        :param columns:
        :return:
        """
        for entity in self._iter_records(columns, row_filter_query):
            for propertyLoader in self.property_loaders:
                propertyLoader.load_property_value(entity)
            yield entity

    def _iter_records(self, columns: Optional[List[str]], row_filter_query: Optional[str]):
        column_names_as_comma_separated_string = None
        if columns is not None:
            column_names_as_comma_separated_string = ",".join(columns)
        return self.table_service.query_entities(self.table_name, select=column_names_as_comma_separated_string,
                filter=row_filter_query)

    def insert_data_frame_to_table(self, df: pd.DataFrame, partition_key_generator: Callable[[str], str] = None, num_records: int = None):
        """
        Inserts or replace entities of the table corresponding to rows of the DataFrame, where the index of the dataFrame acts as rowKey.
        Values of object type columns in the dataFrame may have to be serialised via json beforehand.
        :param df: DataFrame to be inserted
        :param partition_key_generator: if None, partitionKeys default to tableName
        :param num_records: restrict insertion to first numRecords rows, merely for testing
        """
        for (count, (idx, row)) in enumerate(df.iterrows()):
            if num_records is not None:
                if count >= num_records:
                    break
            entity = row.to_dict()
            entity["RowKey"] = idx
            entity["PartitionKey"] = self.table_name if partition_key_generator is None else partition_key_generator(idx)
            self.insert_or_replace_entity(entity)

    @staticmethod
    def _insert_or_replace_entity_via_batch(entity, batch: TableBatch):
        return batch.insert_or_replace_entity(entity)

    @staticmethod
    def _insert_entity_via_batch(entity, batch: TableBatch):
        return batch.insert_entity(entity)

    def exists(self):
        return self.table_service.exists(self.table_name)


class AzureTablePersistentKeyValueCache(PersistentKeyValueCache):
    """
    PersistentKeyValueCache using Azure Table Storage, see https://docs.microsoft.com/en-gb/azure/storage/tables/
    """
    CACHE_VALUE_IDENTIFIER = "cache_value"

    def __init__(self, table_service: TableService, table_name="cache", partition_key_generator: Callable[[str], str] = None,
            max_batch_size=100, min_size_for_periodic_commit: Optional[int] = 100, deferred_commit_delay_secs=1.0, in_memory=False,
            blob_backend: AzureTableBlobBackend = None, serialiser: Serialiser = None, max_workers: int = None):
        """
        :param table_service: https://docs.microsoft.com/en-us/python/api/azure-cosmosdb-table/azure.cosmosdb.table.tableservice.tableservice?view=azure-python
        :param table_name: name of table, needs to match restrictions for Azure storage resources, see https://docs.microsoft.com/en-gb/azure/azure-resource-manager/management/resource-name-rules
        :param partition_key_generator: callable to generate a partitionKey from provided string, if None partitionKey in requests defaults
            to tableName
        :param max_batch_size: maximal batch size for each commit.
        :param deferred_commit_delay_secs: the time frame during which no new data must be added for a pending transaction to be committed
        :param min_size_for_periodic_commit: minimal size of a batch to be committed in a periodic thread.
            If None, commits are only executed in a deferred manner, i.e. commit only if there is no update for `deferred_commit_delay_secs`
        :param in_memory: boolean flag, to indicate, if table should be loaded in memory at construction
        :param blob_backend: if not None, blob storage will be used to store actual value and cache_value in table only contains a reference
        :param max_workers: maximal number of workers to load data from blob backend
        """

        self._deferredCommitDelaySecs = deferred_commit_delay_secs
        self._partitionKeyGenerator = partition_key_generator

        def create_property_loaders():
            if blob_backend is None and serialiser is None:
                _property_loaders = ()
            elif blob_backend is None and serialiser is not None:
                _property_loaders = (SerialisedPropertyLoader(self.CACHE_VALUE_IDENTIFIER, serialiser),)
            elif blob_backend is not None and serialiser is None:
                property_blob_status_name = self.CACHE_VALUE_IDENTIFIER + "_blob_backed"
                _property_loaders = (BlobBackedPropertyLoader(self.CACHE_VALUE_IDENTIFIER, blob_backend, table_name,
                    property_blob_status_name, max_workers),)
            else:
                property_blob_status_name = self.CACHE_VALUE_IDENTIFIER + "_blob_backed"
                _property_loaders = (BlobBackedSerialisedPropertyLoader(self.CACHE_VALUE_IDENTIFIER, serialiser, blob_backend,
                table_name, property_blob_status_name, max_workers),)
            return _property_loaders

        property_loaders = create_property_loaders()
        self._batch_commit_table = AzureLazyBatchCommitTable(table_name, table_service, property_loaders=property_loaders)
        self._minSizeForPeriodicCommit = min_size_for_periodic_commit
        self._maxBatchSize = max_batch_size
        self._updateHook = PeriodicUpdateHook(deferred_commit_delay_secs, no_update_fn=self._commit, periodic_fn=self._periodically_commit)

        self._in_memory_cache = None

        if in_memory:
            df = self._batch_commit_table.load_table_to_data_frame(columns=['RowKey', self.CACHE_VALUE_IDENTIFIER]).set_index("RowKey")
            _log.info(f"Loaded {len(df)} entries of table {table_name} in memory")
            self._in_memory_cache = df[self.CACHE_VALUE_IDENTIFIER].to_dict()

    def set(self, key, value):
        key_as_string = str(key)
        partition_key = self._get_partition_key_for_row_key(key_as_string)
        entity = {'PartitionKey': partition_key, 'RowKey': key_as_string, self.CACHE_VALUE_IDENTIFIER: value}
        self._batch_commit_table.insert_or_replace_entity(entity)
        self._updateHook.handle_update()

        if self._in_memory_cache is not None:
            self._in_memory_cache[key_as_string] = value

    def get(self, key):
        key_as_string = str(key)
        value = self._get_from_in_memory_cache(key_as_string)
        if value is None:
            value = self._get_from_table(key_as_string)
        return value

    def _get_from_table(self, key: str):
        partition_key = self._get_partition_key_for_row_key(key)
        entity = self._batch_commit_table.get_entity(partition_key, key)
        if entity is not None:
            return entity[self.CACHE_VALUE_IDENTIFIER]
        return None

    def _get_from_in_memory_cache(self, key):
        if self._in_memory_cache is None:
            return None
        return self._in_memory_cache.get(str(key), None)

    def _get_partition_key_for_row_key(self, key: str):
        return self._batch_commit_table.table_name if self._partitionKeyGenerator is None else self._partitionKeyGenerator(key)

    def _commit(self):
        self._batch_commit_table.commit_non_blocking_current_queue_state(self._maxBatchSize)

    def _periodically_commit(self):
        self._batch_commit_table.commit_blocking_largest_partition_from_queue(self._maxBatchSize, self._minSizeForPeriodicCommit)



import re
from typing import Dict, Any, Optional

import mlflow
from matplotlib import pyplot as plt

from .tracking_base import TrackedExperiment, TrackingContext
from .. import VectorModelBase
from ..util import logging


class MLFlowTrackingContext(TrackingContext):
    def __init__(self, name: str, experiment: "MLFlowExperiment", run_id=None, description=""):
        super().__init__(name, experiment)
        if run_id is not None:
            run = mlflow.start_run(run_id)
        else:
            run = mlflow.start_run(run_name=name, description=description)
        self.run = run
        self.log_handler: Optional[logging.MemoryStreamHandler] = None

    @staticmethod
    def _metric_name(name: str):
        result = re.sub(r"\[(.*?)\]", r"_\1", name)  # replace "foo[bar]" with "foo_bar"
        result = re.sub(r"[^a-zA-Z0-9-_. /]+", "_", result)  # replace sequences of unsupported chars with underscore
        return result

    def _track_metrics(self, metrics: Dict[str, float]):
        metrics = {self._metric_name(name): value for name, value in metrics.items()}
        mlflow.log_metrics(metrics)

    def track_figure(self, name: str, fig: plt.Figure):
        mlflow.log_figure(fig, name + ".png")

    def track_text(self, name: str, content: str):
        mlflow.log_text(content, name + ".txt")

    def track_tag(self, tag_name: str, tag_value: str):
        mlflow.set_tag(tag_name, tag_value)

    def _end(self):
        mlflow.end_run()


class MLFlowExperiment(TrackedExperiment[MLFlowTrackingContext]):
    def __init__(self, experiment_name: str, tracking_uri: str, additional_logging_values_dict=None,
            context_prefix: str = "", add_log_to_all_contexts=False):
        """
        :param experiment_name: the name of the experiment, which should be the same for all models of the same kind (i.e. all models evaluated
            under the same conditions)
        :param tracking_uri: the URI of the server (if any); use "" to track in the local file system
        :param additional_logging_values_dict:
        :param context_prefix: a prefix to add to all contexts that are created within the experiment. This can be used to add
            an identifier of a certain execution/run, such that the actual context name passed to `begin_context` can be concise (e.g. just model name).
        :param add_log_to_all_contexts: whether to enable in-memory logging and add the respective log to each context
        """
        mlflow.set_tracking_uri(tracking_uri)
        mlflow.set_experiment(experiment_name=experiment_name)
        super().__init__(context_prefix=context_prefix, additional_logging_values_dict=additional_logging_values_dict)
        self._run_name_to_id = {}
        self.add_log_to_all_contexts = add_log_to_all_contexts

    def _track_values(self, values_dict: Dict[str, Any]):
        with mlflow.start_run():
            mlflow.log_metrics(values_dict)

    def _create_tracking_context(self, name: str, description: str) -> MLFlowTrackingContext:
        run_id = self._run_name_to_id.get(name)
        print(f"create {name}")
        context = MLFlowTrackingContext(name, self, run_id=run_id, description=description)
        self._run_name_to_id[name] = context.run.info.run_id
        return context

    def begin_context_for_model(self, model: VectorModelBase):
        context = super().begin_context_for_model(model)
        if self.add_log_to_all_contexts:
            context.log_handler = logging.add_memory_logger()
        context.track_tag("ModelClass", model.__class__.__name__)
        return context

    def end_context(self, instance: MLFlowTrackingContext):
        print(f"end {instance}")
        if instance.log_handler is not None:
            instance.track_text("log", instance.log_handler.get_log())
        super().end_context(instance)



from azureml.core import Experiment, Workspace
from typing import Dict, Any

from .tracking_base import TrackedExperiment, TContext
from .. import VectorModel
from ..evaluation.evaluator import MetricsDictProvider


class TrackedAzureMLEvaluation:
    """
    Class to automatically track parameters, metrics and artifacts for a single model with azureml-sdk
    """
    def __init__(self, experiment_name: str, workspace: Workspace,
            evaluator: MetricsDictProvider):
        """
        :param experiment_name:
        :param workspace:
        :param evaluator:
        """
        self.experiment_name = experiment_name
        self.evaluator = evaluator
        self.experiment = Experiment(workspace=workspace, name=experiment_name)

    def eval_model(self, model: VectorModel, additional_logging_values_dict: dict = None, **start_logging_kwargs):
        with self.experiment.start_logging(**start_logging_kwargs) as run:
            values_dict = self.evaluator.compute_metrics(model)
            values_dict['str(model)'] = str(model)
            if additional_logging_values_dict is not None:
                values_dict.update(additional_logging_values_dict)
            for name, value in values_dict.items():
                run.log(name, value)


class TrackedAzureMLExperiment(TrackedExperiment):
    def __init__(self, experiment_name: str, workspace: Workspace, additional_logging_values_dict=None):
        """

        :param experiment_name: name of experiment for tracking in workspace
        :param workspace: Azure workspace object
        :param additional_logging_values_dict: additional values to be logged for each run
        """
        self.experiment_name = experiment_name
        self.experiment = Experiment(workspace=workspace, name=experiment_name)
        super().__init__(additional_logging_values_dict=additional_logging_values_dict)

    def _track_values(self, values_dict: Dict[str, Any]):
        with self.experiment.start_logging() as run:
            for name, value in values_dict.items():
                run.log(name, value)

    def _create_tracking_context(self, name: str, description: str) -> TContext:
        raise NotImplementedError()



import logging
import math
import random
from abc import ABC, abstractmethod
from typing import Tuple, Sequence, TypeVar, Generic, Optional, Union

import numpy as np
import pandas as pd
import scipy.stats
from sklearn.model_selection import StratifiedShuffleSplit

from ..util.pickle import setstate
from ..util.string import ToStringMixin

log = logging.getLogger(__name__)

T = TypeVar("T")


class BaseInputOutputData(Generic[T], ABC):
    def __init__(self, inputs: T, outputs: T):
        """
        :param inputs: expected to have shape and __len__
        :param outputs: expected to have shape and __len__
        """
        if len(inputs) != len(outputs):
            raise ValueError("Lengths do not match")
        self.inputs = inputs
        self.outputs = outputs

    def __len__(self):
        return len(self.inputs)

    @abstractmethod
    def filter_indices(self, indices: Sequence[int]) -> __qualname__:
        pass


class InputOutputArrays(BaseInputOutputData[np.ndarray]):
    def __init__(self, inputs: np.ndarray, outputs: np.ndarray):
        super().__init__(inputs, outputs)

    def filter_indices(self, indices: Sequence[int]) -> __qualname__:
        inputs = self.inputs[indices]
        outputs = self.outputs[indices]
        return InputOutputArrays(inputs, outputs)

    def to_torch_data_loader(self, batch_size=64, shuffle=True):
        try:
            import torch
            from torch.utils.data import DataLoader, TensorDataset
        except ImportError:
            raise ImportError(f"Could not import torch, did you install it?")
        dataset = TensorDataset(torch.tensor(self.inputs), torch.tensor(self.outputs))
        return DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)


class InputOutputData(BaseInputOutputData[pd.DataFrame], ToStringMixin):
    """
    Holds input and output data for learning problems
    """
    def __init__(self, inputs: pd.DataFrame, outputs: pd.DataFrame, weights: Optional[Union[pd.Series, "DataPointWeighting"]] = None):
        super().__init__(inputs, outputs)
        if isinstance(weights, DataPointWeighting):
            weights = weights.compute_weights(inputs, outputs)
        self.weights = weights

    def __setstate__(self, state):
        setstate(InputOutputData, self, state, new_optional_properties=["weights"])

    def _tostring_object_info(self) -> str:
        return f"N={len(self.inputs)}, numInputColumns={len(self.inputs.columns)}, numOutputColumns={len(self.outputs.columns)}"

    @classmethod
    def from_data_frame(cls, df: pd.DataFrame, *output_columns: str) -> "InputOutputData":
        """
        :param df: a data frame containing both input and output columns
        :param output_columns: the output column name(s)
        :return: an InputOutputData instance with inputs and outputs separated
        """
        inputs = df[[c for c in df.columns if c not in output_columns]]
        outputs = df[list(output_columns)]
        return cls(inputs, outputs)

    def to_data_frame(self, add_weights: bool = False, weights_col_name: str = "weights") -> pd.DataFrame:
        """
        :param add_weights: whether to add the weights as a column (provided that weights are present)
        :param weights_col_name: the column name to use for weights if `add_weights` is True
        :return: a data frame containing both the inputs and outputs (and optionally the weights)
        """
        df = pd.concat([self.inputs, self.outputs], axis=1)
        if add_weights and self.weights is not None:
            df[weights_col_name] = self.weights
        return df

    def to_df(self, add_weights: bool = False, weights_col_name: str = "weights") -> pd.DataFrame:
        return self.to_data_frame(add_weights=add_weights, weights_col_name=weights_col_name)

    def filter_indices(self, indices: Sequence[int]) -> __qualname__:
        inputs = self.inputs.iloc[indices]
        outputs = self.outputs.iloc[indices]
        weights = None
        if self.weights is not None:
            weights = self.weights.iloc[indices]
        return InputOutputData(inputs, outputs, weights)

    def filter_index(self, index_elements: Sequence[any]) -> __qualname__:
        inputs = self.inputs.loc[index_elements]
        outputs = self.outputs.loc[index_elements]
        weights = None
        if self.weights is not None:
            weights = self.weights
        return InputOutputData(inputs, outputs, weights)

    @property
    def input_dim(self):
        return self.inputs.shape[1]

    @property
    def output_dim(self):
        return self.outputs.shape[1]

    def compute_input_output_correlation(self):
        correlations = {}
        for outputCol in self.outputs.columns:
            correlations[outputCol] = {}
            output_series = self.outputs[outputCol]
            for inputCol in self.inputs.columns:
                input_series = self.inputs[inputCol]
                pcc, pvalue = scipy.stats.pearsonr(input_series, output_series)
                correlations[outputCol][inputCol] = pcc
        return correlations

    def apply_weighting(self, weighting: "DataPointWeighting"):
        self.weights = weighting.compute_weights(self.inputs, self.outputs)


TInputOutputData = TypeVar("TInputOutputData", bound=BaseInputOutputData)


class DataSplitter(ABC, Generic[TInputOutputData]):
    @abstractmethod
    def split(self, data: TInputOutputData) -> Tuple[TInputOutputData, TInputOutputData]:
        pass


class DataSplitterFractional(DataSplitter):
    def __init__(self, fractional_size_of_first_set: float, shuffle=True, random_seed=42):
        if not 0 <= fractional_size_of_first_set <= 1:
            raise Exception(f"invalid fraction: {fractional_size_of_first_set}")
        self.fractionalSizeOfFirstSet = fractional_size_of_first_set
        self.shuffle = shuffle
        self.randomSeed = random_seed

    def split_with_indices(self, data) -> Tuple[Tuple[Sequence[int], Sequence[int]], Tuple[TInputOutputData, TInputOutputData]]:
        num_data_points = len(data)
        split_index = int(num_data_points * self.fractionalSizeOfFirstSet)
        if self.shuffle:
            rand = np.random.RandomState(self.randomSeed)
            indices = rand.permutation(num_data_points)
        else:
            indices = range(num_data_points)
        indices_a = indices[:split_index]
        indices_b = indices[split_index:]
        a = data.filter_indices(list(indices_a))
        b = data.filter_indices(list(indices_b))
        return (indices_a, indices_b), (a, b)

    def split(self, data: TInputOutputData) -> Tuple[TInputOutputData, TInputOutputData]:
        _, (a, b) = self.split_with_indices(data)
        return a, b


class DataSplitterFromDataFrameSplitter(DataSplitter[InputOutputData]):
    """
    Creates a DataSplitter from a DataFrameSplitter, which can be applied either to the input or the output data.
    It supports only InputOutputData, not other subclasses of BaseInputOutputData.
    """
    def __init__(self, data_frame_splitter: "DataFrameSplitter", fractional_size_of_first_set: float, apply_to_input=True):
        """
        :param data_frame_splitter: the splitter to apply
        :param fractional_size_of_first_set: the desired fractional size of the first set when applying the splitter
        :param apply_to_input: if True, apply the splitter to the input data frame; if False, apply it to the output data frame
        """
        self.dataFrameSplitter = data_frame_splitter
        self.fractionalSizeOfFirstSet = fractional_size_of_first_set
        self.applyToInput = apply_to_input

    def split(self, data: InputOutputData) -> Tuple[InputOutputData, InputOutputData]:
        if not isinstance(data, InputOutputData):
            raise ValueError(f"{self} is only applicable to instances of {InputOutputData.__name__}, got {data}")
        df = data.inputs if self.applyToInput else data.outputs
        indices_a, indices_b = self.dataFrameSplitter.compute_split_indices(df, self.fractionalSizeOfFirstSet)
        a = data.filter_indices(list(indices_a))
        b = data.filter_indices(list(indices_b))
        return a, b


class DataSplitterFromSkLearnSplitter(DataSplitter):
    def __init__(self, sklearn_splitter):
        """
        :param sklearn_splitter: an instance of one of the splitter classes from sklearn.model_selection,
            see https://scikit-learn.org/stable/modules/classes.html#module-sklearn.model_selection
        """
        self.sklearn_splitter = sklearn_splitter

    def split(self, data: TInputOutputData) -> Tuple[TInputOutputData, TInputOutputData]:
        splitter_result = self.sklearn_splitter.split(data.inputs, data.outputs)
        split = next(iter(splitter_result))
        first_indices, second_indices = split
        return data.filter_indices(first_indices), data.filter_indices(second_indices)


class DataSplitterStratifiedShuffleSplit(DataSplitterFromSkLearnSplitter):
    def __init__(self, fractional_size_of_first_set: float, random_seed=42):
        super().__init__(StratifiedShuffleSplit(n_splits=1, train_size=fractional_size_of_first_set, random_state=random_seed))

    @staticmethod
    def is_applicable(io_data: InputOutputData):
        class_counts = io_data.outputs.value_counts()
        return all(class_counts >= 2)


class DataFrameSplitter(ABC):
    @abstractmethod
    def compute_split_indices(self, df: pd.DataFrame, fractional_size_of_first_set: float) -> Tuple[Sequence[int], Sequence[int]]:
        pass

    @staticmethod
    def split_with_indices(df: pd.DataFrame, indices_pair: Tuple[Sequence[int], Sequence[int]]) -> Tuple[pd.DataFrame, pd.DataFrame]:
        indices_a, indices_b = indices_pair
        a = df.iloc[indices_a]
        b = df.iloc[indices_b]
        return a, b

    def split(self, df: pd.DataFrame, fractional_size_of_first_set: float) -> Tuple[pd.DataFrame, pd.DataFrame]:
        return self.split_with_indices(df, self.compute_split_indices(df, fractional_size_of_first_set))


class DataFrameSplitterFractional(DataFrameSplitter):
    def __init__(self, shuffle=False, random_seed=42):
        self.randomSeed = random_seed
        self.shuffle = shuffle

    def compute_split_indices(self, df: pd.DataFrame, fractional_size_of_first_set: float) -> Tuple[Sequence[int], Sequence[int]]:
        n = df.shape[0]
        size_a = int(n * fractional_size_of_first_set)
        if self.shuffle:
            rand = np.random.RandomState(self.randomSeed)
            indices = rand.permutation(n)
        else:
            indices = list(range(n))
        indices_a = indices[:size_a]
        indices_b = indices[size_a:]
        return indices_a, indices_b


class DataFrameSplitterColumnEquivalenceClass(DataFrameSplitter):
    """
    Performs a split that keeps together data points/rows that have the same value in a given column, i.e.
    with respect to that column, the items having the same values are viewed as a unit; they form an equivalence class, and all
    data points belonging to the same class are either in the first set or the second set.

    The split is performed at the level of unique items in the column, i.e. the given fraction of equivalence
    classes will end up in the first set and the rest in the second set.

    The list if unique items in the column can be shuffled before applying the split. If no shuffling is applied,
    the original order in the data frame is maintained, and if the items were grouped by equivalence class in the
    original data frame, the split will correspond to a fractional split without shuffling where the split boundary
    is adjusted to not separate an equivalence class.
    """
    def __init__(self, column: str, shuffle=True, random_seed=42):
        """
        :param column: the column which defines the equivalence classes (groups of data points/rows that must not be separated)
        :param shuffle: whether to shuffle the list of unique values in the given column before applying the split
        :param random_seed:
        """
        self.column = column
        self.shuffle = shuffle
        self.random_seed = random_seed

    def compute_split_indices(self, df: pd.DataFrame, fractional_size_of_first_set: float) -> Tuple[Sequence[int], Sequence[int]]:
        values = list(df[self.column].unique())
        if self.shuffle:
            rng = random.Random(self.random_seed)
            rng.shuffle(values)

        num_items_in_first_set = round(fractional_size_of_first_set * len(values))
        first_set_values = set(values[:num_items_in_first_set])

        first_set_indices = []
        second_set_indices = []
        for i, t in enumerate(df.itertuples()):
            if getattr(t, self.column) in first_set_values:
                first_set_indices.append(i)
            else:
                second_set_indices.append(i)
        return first_set_indices, second_set_indices


class DataPointWeighting(ABC):
    @abstractmethod
    def compute_weights(self, x: pd.DataFrame, y: pd.DataFrame) -> pd.Series:
        pass


class DataPointWeightingRegressionTargetIntervalTotalWeight(DataPointWeighting):
    """
    Based on relative weights specified for intervals of the regression target,
    will weight individual data point weights such that the sum of weights of data points within each interval
    satisfies the user-specified relative weight, while ensuring that the total weight of all data points
    is still equal to the number of data points.

    For example, if one specifies `interval_weights` as [(0.5, 1), (inf, 2)], then the data points with target values
    up to 0.5 will get 1/3 of the weight and the remaining data points will get 2/3 of the weight.
    So if there are 100 data points and 50 of them are in the first interval (up to 0.5), then these 50 data points
    will each get weight 1/3*100/50=2/3 and the remaining 50 data points will each get weight 2/3*100/50=4/3.
    The sum of all weights is the number of data points, i.e. 100.

    Example:

    >>> targets = [0.1, 0.2, 0.5, 0.7, 0.8, 0.6]
    >>> x = pd.DataFrame({"foo": np.zeros(len(targets))})
    >>> y = pd.DataFrame({"target": targets})
    >>> weighting = DataPointWeightingRegressionTargetIntervalTotalWeight([(0.5, 1), (1.0, 2)])
    >>> weights = weighting.compute_weights(x, y)
    >>> assert(np.isclose(weights.sum(), len(y)))
    >>> weights.tolist()
    [0.6666666666666666,
     0.6666666666666666,
     0.6666666666666666,
     1.3333333333333333,
     1.3333333333333333,
     1.3333333333333333]
    """
    def __init__(self, intervals_weights: Sequence[Tuple[float, float]]):
        """
        :param intervals_weights: a sequence of tuples (upper_bound, rel_total_weight) where upper_bound is the upper bound
            of the interval, `(lower_bound, upper_bound]`; `lower_bound` is the upper bound of the preceding interval
            or -inf for the first interval. `rel_total_weight` specifies the relative weight of all data points within
            the interval.
        """
        a = -math.inf
        sum_rel_weights = sum(t[1] for t in intervals_weights)
        self.intervals = []
        for b, rel_weight in intervals_weights:
            self.intervals.append(self.Interval(a, b, rel_weight / sum_rel_weights))
            a = b

    class Interval:
        def __init__(self, a: float, b: float, weight_fraction: float):
            self.a = a
            self.b = b
            self.weight_fraction = weight_fraction

        def contains(self, x: float):
            return self.a < x <= self.b

    def compute_weights(self, x: pd.DataFrame, y: pd.DataFrame) -> pd.Series:
        assert len(y.columns) == 1, f"Only a single regression target is supported {self.__class__.__name__}"
        targets = y.iloc[:, 0]
        n = len(x)
        weights = np.zeros(n)
        num_weighted = 0
        for interval in self.intervals:
            mask = np.array([interval.contains(x) for x in targets])
            subset_size = mask.sum()
            num_weighted += subset_size
            weights[mask] = interval.weight_fraction * n / subset_size
        if num_weighted != n:
            raise Exception("Not all data points were weighted. Most likely, the intervals do not cover the entire range of targets")
        return pd.Series(weights, index=x.index)



from datetime import time
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pandas as pd


def ts_next_month(ts: "pd.Timestamp") -> "pd.Timestamp":
    m = ts.month
    if m == 12:
        return ts.replace(year=ts.year+1, month=1)
    else:
        return ts.replace(month=m+1)


def time_of_day(ts: "pd.Timestamp") -> float:
    """
    :param ts: the timestamp
    :return: the time of day as a floating point number in [0, 24)
    """
    return ts.hour + ts.minute / 60


class TimeInterval:
    def __init__(self, start: "pd.Timestamp", end: "pd.Timestamp"):
        self.start = start
        self.end = end

    def contains(self, t: "pd.Timestamp"):
        return self.start <= t <= self.end

    def contains_time(self, t: time):
        """
        :param t: a time of day
        :return: True iff the time interval contains the given time of day at least once, False otherwise
        """
        if (self.end - self.start).total_seconds() >= (60 * 60 * 24):
            return True
        return self.contains(self.start.replace(hour=t.hour, minute=t.minute, second=t.second, microsecond=t.microsecond)) or \
            self.contains(self.end.replace(hour=t.hour, minute=t.minute, second=t.second, microsecond=t.microsecond))

    def overlaps_with(self, other: "TimeInterval") -> bool:
        other_ends_before = other.end <= self.start
        other_starts_after = other.start >= self.end
        return not (other_ends_before or other_starts_after)

    def intersection(self, other: "TimeInterval") -> "TimeInterval":
        return TimeInterval(max(self.start, other.start), min(self.end, other.end))

    def time_delta(self) -> "pd.Timedelta":
        return self.end - self.start

    def mid_timestamp(self) -> "pd.Timestamp":
        midTime: pd.Timestamp = self.start + 0.5 * self.time_delta()
        return midTime


def format_duration(seconds: float):
    if seconds < 60:
        return f"{seconds:.1f} seconds"
    elif seconds < 3600:
        minutes, secs = divmod(seconds, 60)
        return f"{int(minutes)} minutes, {secs:.1f} seconds"
    else:
        hours, remainder = divmod(seconds, 3600)
        minutes, secs = divmod(remainder, 60)
        return f"{int(hours)} hours, {int(minutes)} minutes, {secs:.1f} seconds"



from typing import Sequence, Any

import numpy as np


class ValueOneHotEncoder:
    """
    A simple one-hot encoder, which encodes individual values.
    A one-hot encoder transforms a categorical input value into an array whose length is the number of categories where all values
    are zero except one whose value is one, indicating the category that is active.
    """
    def __init__(self, ignore_unknown=True):
        """
        :param ignore_unknown: whether unknown input values (not seen during fit) shall be ignored, resulting in an array of zeroes;
            if False, throw an exception instead
        """
        self.categories = None
        self.category2index = None
        self.ignoreUnknown = ignore_unknown

    def fit(self, values: Sequence[Any]):
        unique_values = np.unique(values)
        self.categories = sorted(unique_values)
        self.category2index = {category: idx for idx, category in enumerate(self.categories)}

    def transform(self, value) -> np.ndarray:
        a = np.zeros(len(self.categories))
        category_idx = self.category2index.get(value)
        if category_idx is None:
            if not self.ignoreUnknown:
                raise Exception(f"Got unknown value '{value}'")
        else:
            a[category_idx] = 1.0
        return a
